"""Utilities for calculating the labware origin offset position."""

import dataclasses
import enum
from typing import Optional, Union, overload

from typing_extensions import assert_type

from opentrons_shared_data.deck.types import DeckDefinitionV5, SlotDefV3
from opentrons_shared_data.labware.labware_definition import (
    AxisAlignedBoundingBox3D,
    Extents,
    LabwareDefinition,
    LabwareDefinition2,
    LabwareDefinition3,
    Vector3D,
)
from opentrons_shared_data.labware.types import (
    LocatingFeatures,
    SlotFootprintAsChildFeature,
    SlotFootprintAsParentFeature,
    SpringDirectionalForce,
)
from opentrons_shared_data.labware.types import (
    Vector3D as LabwareVector3D,
)
from opentrons_shared_data.module.types import ModuleOrientation

from .errors import (
    IncompatibleLocatingFeatureError,
    InvalidLabwarePlacementError,
    MissingLocatingFeatureError,
)
from opentrons.protocol_engine.errors import (
    InvalidModuleOrientation,
    LabwareNotOnDeckError,
    LabwareOffsetDoesNotExistError,
)
from opentrons.protocol_engine.resources.labware_validation import (
    is_absorbance_reader_lid,
    validate_definition_is_lid,
)
from opentrons.protocol_engine.types import (
    WASTE_CHUTE_LOCATION,
    AddressableArea,
    AddressableAreaLocation,
    DeckLocationDefinition,
    DeckSlotLocation,
    LabwareLocation,
    LabwareMovementOffsetData,
    LabwareOffsetVector,
    ModuleDefinition,
    ModuleLocation,
    ModuleModel,
    OnLabwareLocation,
)
from opentrons.types import DeckSlotName, Point

_OFFSET_ON_TC_OT2 = Point(x=0, y=0, z=10.7)

LabwareStackupAncestorDefinition = Union[
    DeckLocationDefinition,
    ModuleDefinition,
]
_LabwareStackupDefinition = Union[
    DeckLocationDefinition, ModuleDefinition, LabwareDefinition
]


class LabwareOriginContext(enum.Enum):
    """Context for labware origin calculations."""

    PIPETTING = enum.auto()
    GRIPPER_PICKING_UP = enum.auto()
    GRIPPER_DROPPING = enum.auto()


@dataclasses.dataclass
class _Labware3SupportedParentDefinition:
    features: LocatingFeatures
    extents: Extents


@dataclasses.dataclass
class _GripperOffsets:
    pick_up_offset: Point
    drop_offset: Point


def get_stackup_origin_to_labware_origin(
    context: LabwareOriginContext,
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_name: DeckSlotName,
    module_parent_to_child_offset: Point | None,
    deck_definition: DeckDefinitionV5,
) -> Point:
    """Returns the offset from the stackup placement origin to child labware origin.

    Accounts for offset differences caused by context.
    """
    if context == LabwareOriginContext.PIPETTING:
        return _get_stackup_origin_to_lw_origin(
            stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
            underlying_ancestor_definition=underlying_ancestor_definition,
            module_parent_to_child_offset=module_parent_to_child_offset,
            deck_definition=deck_definition,
            slot_name=slot_name,
        )
    else:
        gripper_offsets = _total_nominal_gripper_offsets(
            stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
            underlying_ancestor_definition=underlying_ancestor_definition,
            slot_name=slot_name,
            deck_definition=deck_definition,
        )
        gripper_offset = (
            gripper_offsets.pick_up_offset
            if context == LabwareOriginContext.GRIPPER_PICKING_UP
            else gripper_offsets.drop_offset
        )

        return gripper_offset + _get_stackup_origin_to_lw_origin(
            stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
            underlying_ancestor_definition=underlying_ancestor_definition,
            module_parent_to_child_offset=module_parent_to_child_offset,
            deck_definition=deck_definition,
            slot_name=slot_name,
        )


def _get_stackup_origin_to_lw_origin(
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    module_parent_to_child_offset: Point | None,
    deck_definition: DeckDefinitionV5,
    slot_name: DeckSlotName,
    is_topmost_labware: bool = True,
) -> Point:
    """Returns the offset from the stackup placement origin to child labware origin."""
    definition, location = stackup_lw_info_top_to_bottom[0]
    underlying_ancestor_orientation = _get_underlying_ancestor_orientation(
        underlying_ancestor_definition, slot_name
    )

    if isinstance(
        location, (AddressableAreaLocation, DeckSlotLocation, ModuleLocation)
    ):
        return _get_parent_placement_origin_to_lw_origin_by_location(
            labware_location=location,
            labware_definition=definition,
            parent_definition=underlying_ancestor_definition,
            deck_definition=deck_definition,
            module_parent_to_child_offset=module_parent_to_child_offset,
            is_topmost_labware=is_topmost_labware,
            underlying_ancestor_orientation=underlying_ancestor_orientation,
        )
    elif isinstance(location, OnLabwareLocation):
        parent_definition = stackup_lw_info_top_to_bottom[1][0]

        parent_placement_origin_to_lw_origin = (
            _get_parent_placement_origin_to_lw_origin_by_location(
                labware_location=location,
                labware_definition=definition,
                parent_definition=parent_definition,
                deck_definition=deck_definition,
                module_parent_to_child_offset=module_parent_to_child_offset,
                is_topmost_labware=is_topmost_labware,
                underlying_ancestor_orientation=underlying_ancestor_orientation,
            )
        )
        remaining_lw_defs_locs_top_to_bottom = stackup_lw_info_top_to_bottom[1:]

        return parent_placement_origin_to_lw_origin + _get_stackup_origin_to_lw_origin(
            stackup_lw_info_top_to_bottom=remaining_lw_defs_locs_top_to_bottom,
            underlying_ancestor_definition=underlying_ancestor_definition,
            module_parent_to_child_offset=module_parent_to_child_offset,
            deck_definition=deck_definition,
            slot_name=slot_name,
            is_topmost_labware=False,
        )
    elif location == WASTE_CHUTE_LOCATION:
        raise LabwareNotOnDeckError(
            f"Cannot access {definition.metadata.displayName} because it is in the waste chute."
        )
    else:
        raise LabwareNotOnDeckError(
            f"Cannot access {definition.metadata.displayName} since it is not on the deck. "
            "Either it has been loaded off-deck or its been moved off-deck."
        )


def _get_underlying_ancestor_orientation(
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_id: DeckSlotName,
) -> ModuleOrientation:
    if isinstance(underlying_ancestor_definition, ModuleDefinition):
        orientation = underlying_ancestor_definition.orientation.get(slot_id.id)
        if orientation == "left":
            return ModuleOrientation.LEFT
        elif orientation == "right":
            return ModuleOrientation.RIGHT
        elif orientation == "center":
            return ModuleOrientation.CENTER
        else:
            raise InvalidModuleOrientation(
                f"Module {underlying_ancestor_definition.moduleType} does "
                f"not contain a valid orientation for slot {slot_id}."
            )

    elif isinstance(underlying_ancestor_definition, AddressableArea):
        return underlying_ancestor_definition.orientation
    else:
        return underlying_ancestor_definition["orientation"]


def _get_parent_placement_origin_to_lw_origin_by_location(
    labware_location: LabwareLocation,
    labware_definition: LabwareDefinition,
    parent_definition: _LabwareStackupDefinition,
    deck_definition: DeckDefinitionV5,
    module_parent_to_child_offset: Point | None,
    underlying_ancestor_orientation: ModuleOrientation,
    is_topmost_labware: bool,
) -> Point:
    if isinstance(labware_location, ModuleLocation):
        if module_parent_to_child_offset is None:
            raise ValueError(
                "Expected value for module_parent_to_child_offset, received None."
            )
        else:
            return _get_parent_placement_origin_to_lw_origin(
                child_labware=labware_definition,
                parent_deck_item=parent_definition,  # type: ignore[arg-type]
                module_parent_to_child_offset=module_parent_to_child_offset,
                deck_definition=deck_definition,
                is_topmost_labware=is_topmost_labware,
                labware_location=labware_location,
                underlying_ancestor_orientation=underlying_ancestor_orientation,
            )
    elif isinstance(labware_location, OnLabwareLocation):
        return _get_parent_placement_origin_to_lw_origin(
            child_labware=labware_definition,
            parent_deck_item=parent_definition,  # type: ignore[arg-type]
            module_parent_to_child_offset=None,
            deck_definition=deck_definition,
            is_topmost_labware=is_topmost_labware,
            labware_location=labware_location,
            underlying_ancestor_orientation=underlying_ancestor_orientation,
        )
    elif isinstance(labware_location, (DeckSlotLocation, AddressableAreaLocation)):
        return _get_parent_placement_origin_to_lw_origin(
            child_labware=labware_definition,
            parent_deck_item=parent_definition,  # type: ignore[arg-type]
            module_parent_to_child_offset=None,
            deck_definition=deck_definition,
            is_topmost_labware=is_topmost_labware,
            labware_location=labware_location,
            underlying_ancestor_orientation=underlying_ancestor_orientation,
        )
    else:
        raise ValueError(f"Invalid labware location: {labware_location}")


@overload
def _get_parent_placement_origin_to_lw_origin(
    child_labware: LabwareDefinition,
    parent_deck_item: ModuleDefinition,
    module_parent_to_child_offset: Point,
    deck_definition: DeckDefinitionV5,
    is_topmost_labware: bool,
    labware_location: ModuleLocation,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point: ...


@overload
def _get_parent_placement_origin_to_lw_origin(
    child_labware: LabwareDefinition,
    parent_deck_item: DeckLocationDefinition,
    module_parent_to_child_offset: None,
    deck_definition: DeckDefinitionV5,
    is_topmost_labware: bool,
    labware_location: Union[DeckSlotLocation, AddressableAreaLocation],
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point: ...


@overload
def _get_parent_placement_origin_to_lw_origin(
    child_labware: LabwareDefinition,
    parent_deck_item: LabwareDefinition,
    module_parent_to_child_offset: None,
    deck_definition: DeckDefinitionV5,
    is_topmost_labware: bool,
    labware_location: OnLabwareLocation,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point: ...


def _get_parent_placement_origin_to_lw_origin(
    child_labware: LabwareDefinition,
    parent_deck_item: _LabwareStackupDefinition,
    module_parent_to_child_offset: Point | None,
    deck_definition: DeckDefinitionV5,
    is_topmost_labware: bool,
    labware_location: LabwareLocation,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point:
    """Returns the offset from parent entity's placement origin to child labware origin.

    Placement origin varies depending on the parent entity type (labware v3 are the back left bottom, and
    labware v2, modules, & deck location types are the front left bottom).

    Only parent-child specific offsets are calculated. Offsets that apply to a single entity
    (ex., module cal) or the entire stackup (ex., LPC) are handled elsewhere.
    """
    if isinstance(child_labware, LabwareDefinition2) or isinstance(
        parent_deck_item, LabwareDefinition2
    ):
        parent_deck_item_origin_to_child_labware_placement_origin = (
            _get_parent_deck_item_origin_to_child_labware_placement_origin(
                child_labware=child_labware,
                parent_deck_item=parent_deck_item,
                module_parent_to_child_offset=module_parent_to_child_offset,
                deck_definition=deck_definition,
                labware_location=labware_location,
            )
        )

        # For v2 definitions, cornerOffsetFromSlot is the parent entity placement origin to child labware origin offset.
        # For compatibility with historical (buggy?) behavior,
        # we only consider it when the child labware is the topmost labware in a stackup.
        if isinstance(child_labware, LabwareDefinition2):
            parent_deck_item_to_child_labware_offset = (
                Point.from_xyz_attrs(child_labware.cornerOffsetFromSlot)
                if is_topmost_labware
                else Point(0, 0, 0)
            )

            return (
                parent_deck_item_origin_to_child_labware_placement_origin
                + parent_deck_item_to_child_labware_offset
            )
        else:
            assert isinstance(child_labware, LabwareDefinition3)
            parent_deck_item_to_child_labware_back_left = Point(
                x=0, y=child_labware.extents.total.frontRightTop.y * -1, z=0
            )
            child_labware_back_left_to_child_labware_origin = (
                _get_corner_offset_from_extents(child_labware)
                if is_topmost_labware
                else Point(0, 0, 0)
            )

            return (
                parent_deck_item_origin_to_child_labware_placement_origin  # Only the Z-offset in this case.
                + parent_deck_item_to_child_labware_back_left
                + child_labware_back_left_to_child_labware_origin
            )
    else:
        # For v3 definitions, get the vector from the back left bottom to the front right bottom.
        assert_type(child_labware, LabwareDefinition3)

        # TODO(jh, 06-25-25): This code is entirely temporary and only exists for the purposes of more useful
        #  snapshot testing. This code should exist in NO capacity after features are implemented outside of the
        #  module_parent_to_child_offset.
        if _shim_does_locating_feature_pair_exist(
            child_labware=child_labware,
            parent_deck_item=_get_standardized_parent_deck_item(parent_deck_item),
        ):
            parent_deck_item_origin_to_child_labware_placement_origin = (
                _module_parent_to_child_offset(
                    module_parent_to_child_offset, labware_location
                )
            )
        else:
            parent_deck_item_origin_to_child_labware_placement_origin = (
                _get_parent_deck_item_origin_to_child_labware_placement_origin(
                    child_labware=child_labware,
                    parent_deck_item=parent_deck_item,
                    module_parent_to_child_offset=module_parent_to_child_offset,
                    deck_definition=deck_definition,
                    labware_location=labware_location,
                )
            )

        parent_deck_item_to_child_labware_feature_offset = (
            _parent_deck_item_to_child_labware_feature_offset(
                child_labware=child_labware,
                parent_deck_item=_get_standardized_parent_deck_item(parent_deck_item),
                underlying_ancestor_orientation=underlying_ancestor_orientation,
            )
        ) + _feature_exception_offsets(
            deck_definition=deck_definition, parent_deck_item=parent_deck_item
        )

        return (
            parent_deck_item_origin_to_child_labware_placement_origin
            + parent_deck_item_to_child_labware_feature_offset
        )


def _get_parent_deck_item_origin_to_child_labware_placement_origin(
    child_labware: LabwareDefinition,
    parent_deck_item: _LabwareStackupDefinition,
    module_parent_to_child_offset: Point | None,
    deck_definition: DeckDefinitionV5,
    labware_location: LabwareLocation,
) -> Point:
    """Get the offset vector from parent entity origin to child labware placement origin."""
    if isinstance(labware_location, (DeckSlotLocation, AddressableAreaLocation)):
        return Point(x=0, y=0, z=0)

    elif isinstance(labware_location, ModuleLocation):
        assert isinstance(parent_deck_item, ModuleDefinition)

        child_labware_overlap_with_parent_deck_item = (
            _get_child_labware_overlap_with_parent_module(
                child_labware=child_labware,
                parent_module_model=parent_deck_item.model,
                deck_definition=deck_definition,
            )
        )
        module_parent_to_child_offset = _module_parent_to_child_offset(
            module_parent_to_child_offset, labware_location
        )

        return (
            module_parent_to_child_offset - child_labware_overlap_with_parent_deck_item
        )

    elif isinstance(labware_location, OnLabwareLocation):
        assert isinstance(parent_deck_item, (LabwareDefinition2, LabwareDefinition3))

        # TODO(jh, 06-05-25): This logic is slightly duplicative of LabwareView get_dimensions. Can we unify?
        if isinstance(parent_deck_item, LabwareDefinition2):
            parent_deck_item_height = parent_deck_item.dimensions.zDimension
        else:
            assert_type(parent_deck_item, LabwareDefinition3)
            parent_deck_item_height = (
                parent_deck_item.extents.total.frontRightTop.z
                - parent_deck_item.extents.total.backLeftBottom.z
            )

        child_labware_overlap_with_parent_deck_item = (
            _get_child_labware_overlap_with_parent_labware(
                child_labware=child_labware,
                parent_labware_name=parent_deck_item.parameters.loadName,
            )
        )

        return Point(
            x=child_labware_overlap_with_parent_deck_item.x,
            y=child_labware_overlap_with_parent_deck_item.y,
            z=parent_deck_item_height - child_labware_overlap_with_parent_deck_item.z,
        )

    else:
        raise TypeError(f"Unsupported labware location type: {labware_location}")


def _get_corner_offset_from_extents(child_labware: LabwareDefinition3) -> Point:
    """Derive the corner offset from slot from a LabwareDefinition3's extents."""
    back_left_bottom = child_labware.extents.total.backLeftBottom

    x = back_left_bottom.x
    y = back_left_bottom.y * -1
    z = back_left_bottom.z

    return Point(x, y, z)


def _module_parent_to_child_offset(
    module_parent_to_child_offset: Point | None,
    labware_location: LabwareLocation,
) -> Point:
    """Returns the module offset if applicable."""
    if (
        isinstance(labware_location, ModuleLocation)
        and module_parent_to_child_offset is not None
    ):
        return Point.from_xyz_attrs(module_parent_to_child_offset)
    else:
        return Point(0, 0, 0)


def _shim_does_locating_feature_pair_exist(
    child_labware: LabwareDefinition3,
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> bool:
    """Temporary util."""
    slot_footprint_exists = (
        parent_deck_item.features.get("slotFootprintAsParent") is not None
        and child_labware.features.get("slotFootprintAsChild") is not None
    )
    flex_tiprack_lid_exists = (
        parent_deck_item.features.get("opentronsFlexTipRackLidAsParent") is not None
        and child_labware.features.get("opentronsFlexTipRackLidAsChild") is not None
    )
    hs_universal_flat_adapter_exists = (
        parent_deck_item.features.get("heaterShakerUniversalFlatAdapter") is not None
        and child_labware.features.get("flatSupportThermalCouplingAsChild") is not None
    )
    hs_universal_flat_adapter_screw_anchored_exists = (
        parent_deck_item.features.get("screwAnchoredAsParent") is not None
        and child_labware.features.get("heaterShakerUniversalFlatAdapter") is not None
    )
    screw_anchored_exists = (
        parent_deck_item.features.get("screwAnchoredAsParent") is not None
        and child_labware.features.get("screwAnchoredAsChild") is not None
    )

    return (
        slot_footprint_exists
        or flex_tiprack_lid_exists
        or hs_universal_flat_adapter_exists
        or hs_universal_flat_adapter_screw_anchored_exists
        or screw_anchored_exists
    )


def _get_standardized_parent_deck_item(
    parent_deck_item: Union[
        LabwareDefinition3, DeckLocationDefinition, ModuleDefinition
    ],
) -> _Labware3SupportedParentDefinition:
    """Returns a standardized parent deck item interface."""
    if isinstance(parent_deck_item, ModuleDefinition):
        slot_footprint_as_parent = _module_slot_footprint_as_parent(parent_deck_item)
        if slot_footprint_as_parent is not None:
            return _Labware3SupportedParentDefinition(
                features={
                    **parent_deck_item.features,
                    "slotFootprintAsParent": slot_footprint_as_parent,
                },
                extents=parent_deck_item.extents,
            )
        else:
            return _Labware3SupportedParentDefinition(
                features=parent_deck_item.features, extents=parent_deck_item.extents
            )
    elif isinstance(parent_deck_item, AddressableArea):
        extents = Extents(
            total=AxisAlignedBoundingBox3D(
                backLeftBottom=Vector3D(x=0, y=0, z=0),
                frontRightTop=Vector3D(
                    x=parent_deck_item.bounding_box.x,
                    y=parent_deck_item.bounding_box.y * 1,
                    z=parent_deck_item.bounding_box.z,
                ),
            )
        )

        slot_footprint_as_parent = _aa_slot_footprint_as_parent(parent_deck_item)
        if slot_footprint_as_parent is not None:
            return _Labware3SupportedParentDefinition(
                features={
                    **parent_deck_item.features,
                    "slotFootprintAsParent": slot_footprint_as_parent,
                },
                extents=extents,
            )
        else:
            return _Labware3SupportedParentDefinition(
                parent_deck_item.features, extents=extents
            )
    elif isinstance(parent_deck_item, LabwareDefinition3):
        return _Labware3SupportedParentDefinition(
            features=parent_deck_item.features, extents=parent_deck_item.extents
        )
    # The slotDefV3 case.
    else:
        extents = Extents(
            total=AxisAlignedBoundingBox3D(
                backLeftBottom=Vector3D(x=0, y=0, z=0),
                frontRightTop=Vector3D(
                    x=parent_deck_item["boundingBox"]["xDimension"],
                    y=parent_deck_item["boundingBox"]["yDimension"] * 1,
                    z=parent_deck_item["boundingBox"]["zDimension"],
                ),
            )
        )
        slot_footprint_as_parent = _slot_def_slot_footprint_as_parent(parent_deck_item)
        return _Labware3SupportedParentDefinition(
            features={
                **parent_deck_item["features"],
                "slotFootprintAsParent": slot_footprint_as_parent,
            },
            extents=extents,
        )


def _module_slot_footprint_as_parent(
    parent_deck_item: ModuleDefinition,
) -> SlotFootprintAsParentFeature | None:
    """Returns the slot footprint as parent feature if inherently supported by the module definition.

    This utility is a normalization shim until labwareOffset + labwareInterfaceX/YDimension is deleted in module defs
    and replaced with the same slotFootprintAsParent that exists in labware def v3.
    """
    dimensions = parent_deck_item.dimensions
    if (
        dimensions.labwareInterfaceYDimension is None
        or dimensions.labwareInterfaceXDimension is None
    ):
        return None
    else:
        # Modules with springs would require special mating types and therefore are not handled here.
        return SlotFootprintAsParentFeature(
            z=0,
            backLeft={"x": 0, "y": dimensions.labwareInterfaceYDimension},
            frontRight={"x": dimensions.labwareInterfaceXDimension, "y": 0},
        )


def _aa_slot_footprint_as_parent(
    parent_deck_item: AddressableArea,
) -> SlotFootprintAsParentFeature | None:
    """Returns the slot footprint as parent feature for addressable areas.

    This utility is a normalization shim until bounding box in deck defs and
    replaced with the same slotFootprintAsParent that exists in labware def v3.
    """
    bb = parent_deck_item.bounding_box

    if parent_deck_item.mating_surface_unit_vector is not None:
        if parent_deck_item.mating_surface_unit_vector == [-1, 1, -1]:
            return SlotFootprintAsParentFeature(
                z=0,
                backLeft={"x": 0, "y": bb.y},
                frontRight={"x": bb.x, "y": 0},
                springDirectionalForce="backLeftBottom",
            )
        else:
            raise NotImplementedError(
                "Slot footprint as parent does not support mating surface unit vector."
            )
    else:
        return SlotFootprintAsParentFeature(
            z=0,
            backLeft={"x": 0, "y": bb.y},
            frontRight={"x": bb.x, "y": 0},
        )


def _slot_def_slot_footprint_as_parent(
    parent_deck_item: SlotDefV3,
) -> SlotFootprintAsParentFeature:
    """Returns the slot footprint as parent feature for slot definitions.

    This utility is a normalization shim until bounding box in deck defs and
    replaced with the same slotFootprintAsParent that exists in labware def v3.
    """
    bb = parent_deck_item["boundingBox"]
    return SlotFootprintAsParentFeature(
        z=0,
        backLeft={"x": 0, "y": bb["yDimension"]},
        frontRight={"x": bb["xDimension"], "y": 0},
        springDirectionalForce="backLeftBottom",
    )


def _parent_deck_item_to_child_labware_feature_offset(
    child_labware: LabwareDefinition3,
    parent_deck_item: _Labware3SupportedParentDefinition,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point:
    """Get the offset vector from the parent entity origin to the child labware origin."""
    if parent_deck_item.features.get("heaterShakerUniversalFlatAdapter") is not None:
        if child_labware.features.get("flatSupportThermalCouplingAsChild") is not None:
            return _parent_origin_to_heater_shaker_universal_flat_adapter_feature(
                parent_deck_item=parent_deck_item,
                underlying_ancestor_orientation=underlying_ancestor_orientation,
            ) + _heater_shaker_universal_flat_adapter_feature_to_child_origin(
                child_labware=child_labware,
                underlying_ancestor_orientation=underlying_ancestor_orientation,
            )
        else:
            raise MissingLocatingFeatureError(
                labware_name=child_labware.metadata.displayName,
                required_feature="flatSupportThermalCouplingAsChild",
            )

    elif (
        parent_deck_item.features.get("opentronsFlexTipRackLidAsParent") is not None
        and child_labware.features.get("opentronsFlexTipRackLidAsChild") is not None
    ):
        # TODO(jh, 07-29-25): Support center X/Y calculation after addressing grip point
        # calculations. See #18929 discussion.
        return _parent_origin_to_flex_tip_rack_lid_feature(
            parent_deck_item
        ) + _flex_tip_rack_lid_feature_to_child_origin(child_labware)
    elif (
        parent_deck_item.features.get("screwAnchoredAsParent") is not None
        and _get_screw_anchored_center_for_child(
            child_labware, underlying_ancestor_orientation
        )
        is not None
    ):
        return _parent_origin_to_screw_anchored_feature(
            parent_deck_item
        ) + _screw_anchored_feature_to_child_origin(
            child_labware, underlying_ancestor_orientation
        )
    elif (
        parent_deck_item.features.get("slotFootprintAsParent") is not None
        and child_labware.features.get("slotFootprintAsChild") is not None
    ):
        spring_force = _get_spring_force(child_labware, parent_deck_item)

        if spring_force is not None:
            if spring_force == "backLeftBottom":
                return _parent_origin_to_slot_back_left_bottom(
                    parent_deck_item
                ) + _slot_back_left_bottom_to_child_origin(child_labware)
            else:
                raise NotImplementedError(f"Spring force: {spring_force}")
        else:
            return _parent_origin_to_slot_bottom_center(
                parent_deck_item
            ) + slot_bottom_center_to_child_origin(child_labware)
    else:
        # TODO(jh, 06-25-25): This is a temporary shim to unblock FE usage with LW Def3 and more accurately diff
        #  ongoing positioning snapshot changes, but we should throw an error  after adding all locating features
        #  if no appropriate LF pair is found.
        return Point(0, 0, 0)


def _get_spring_force(
    child_labware: LabwareDefinition3,
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> SpringDirectionalForce | None:
    """Returns whether the parent-child stackup has a spring that affects positioning."""
    assert parent_deck_item.features.get("slotFootprintAsParent") is not None
    assert child_labware.features.get("slotFootprintAsChild") is not None

    parent_spring_force = parent_deck_item.features["slotFootprintAsParent"].get(
        "springDirectionalForce"
    )
    child_spring_force = child_labware.features["slotFootprintAsChild"].get(
        "springDirectionalForce"
    )

    if parent_spring_force is not None and child_spring_force is not None:
        if parent_spring_force != child_spring_force:
            raise IncompatibleLocatingFeatureError(
                parent_feature=f"slotFootprintAsParent spring force: {parent_spring_force}",
                child_feature=f"slotFootprintAsChild spring force: {child_spring_force}",
            )

    return parent_spring_force or child_spring_force


def _get_screw_anchored_center_for_child(
    child_labware: LabwareDefinition3,
    underlying_ancestor_orientation: ModuleOrientation,
) -> LabwareVector3D | None:
    """Returns the screw center if it exists in any locating feature that supports screw anchoring."""
    hs_universal_flat_adapter_feature = child_labware.features.get(
        "heaterShakerUniversalFlatAdapter"
    )
    screw_anchored_as_child_feature = child_labware.features.get("screwAnchoredAsChild")

    if hs_universal_flat_adapter_feature is not None:
        if underlying_ancestor_orientation == ModuleOrientation.LEFT:
            x = hs_universal_flat_adapter_feature["deckLeft"]["screwCenter"]["x"]
            y = hs_universal_flat_adapter_feature["deckLeft"]["screwCenter"]["y"]
            return LabwareVector3D(x=x, y=y, z=0)
        elif underlying_ancestor_orientation == ModuleOrientation.RIGHT:
            x = hs_universal_flat_adapter_feature["deckRight"]["screwCenter"]["x"]
            y = hs_universal_flat_adapter_feature["deckRight"]["screwCenter"]["y"]
            return LabwareVector3D(x=x, y=y, z=0)
        else:
            raise InvalidLabwarePlacementError(
                feature_name="heaterShakerUniversalFlatAdapter",
                invalid_placement=ModuleOrientation.CENTER.value,
            )
    elif screw_anchored_as_child_feature is not None:
        return screw_anchored_as_child_feature["screwCenter"]
    else:
        return None


def _parent_origin_to_flex_tip_rack_lid_feature(
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> Point:
    """Returns the offset from a deck item's origin to the Flex tip rack lid locating feature."""
    flex_tip_rack_lid_as_parent = parent_deck_item.features.get(
        "opentronsFlexTipRackLidAsParent"
    )
    assert flex_tip_rack_lid_as_parent is not None

    return Point(x=0, y=0, z=flex_tip_rack_lid_as_parent["matingZ"])


def _parent_origin_to_slot_bottom_center(
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> Point:
    """Returns the offset from a deck item's origin to the bottom center of the slot that it provides."""
    slot_footprint_as_parent = parent_deck_item.features.get("slotFootprintAsParent")
    assert slot_footprint_as_parent is not None

    x = (
        slot_footprint_as_parent["frontRight"]["x"]
        + slot_footprint_as_parent["backLeft"]["x"]
    ) / 2
    y = (
        slot_footprint_as_parent["frontRight"]["y"]
        + slot_footprint_as_parent["backLeft"]["y"]
    ) / 2
    z = slot_footprint_as_parent["z"]

    return Point(x, y, z)


def _parent_origin_to_slot_back_left_bottom(
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> Point:
    """Returns the offset from a deck item's origin to the back left bottom of the slot that it provides."""
    slot_footprint_as_parent = parent_deck_item.features.get("slotFootprintAsParent")
    assert slot_footprint_as_parent is not None

    x = slot_footprint_as_parent["backLeft"]["x"]
    y = slot_footprint_as_parent["backLeft"]["y"]
    z = slot_footprint_as_parent["z"]

    return Point(x, y, z)


def _parent_origin_to_heater_shaker_universal_flat_adapter_feature(
    parent_deck_item: _Labware3SupportedParentDefinition,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point:
    """Returns the offset from a deck item's origin to the Heater Shaker Universal Flat Adapter locating feature."""
    flat_adapter_feature = parent_deck_item.features.get(
        "heaterShakerUniversalFlatAdapter"
    )
    assert flat_adapter_feature is not None

    flat_well_support_z = flat_adapter_feature["flatSupportThermalCouplingZ"]
    extents = parent_deck_item.extents.total

    if underlying_ancestor_orientation == ModuleOrientation.LEFT:
        left_wall_x = flat_adapter_feature["deckLeft"]["wallX"]
        left_side_center_x = extents.backLeftBottom.x + left_wall_x
        left_side_center_y = (extents.backLeftBottom.y + extents.frontRightTop.y) / 2

        return Point(left_side_center_x, left_side_center_y, flat_well_support_z)
    elif underlying_ancestor_orientation == ModuleOrientation.RIGHT:
        right_wall_x = flat_adapter_feature["deckRight"]["wallX"]
        right_side_center_x = extents.frontRightTop.x + right_wall_x
        right_side_center_y = (extents.backLeftBottom.y + extents.frontRightTop.y) / 2

        return Point(right_side_center_x, right_side_center_y, flat_well_support_z)

    else:
        raise InvalidLabwarePlacementError(
            feature_name="heaterShakerUniversalFlatAdapter",
            invalid_placement=ModuleOrientation.CENTER.value,
        )


def _parent_origin_to_screw_anchored_feature(
    parent_deck_item: _Labware3SupportedParentDefinition,
) -> Point:
    """Returns the offset from a deck item's origin to the `screwAnchoredAsParent` locating feature."""
    feature = parent_deck_item.features.get("screwAnchoredAsParent")
    assert feature is not None

    screw_center_x = feature["screwCenter"]["x"]
    screw_center_y = feature["screwCenter"]["y"]
    screw_center_z = feature["screwCenter"]["z"]

    return Point(x=screw_center_x, y=screw_center_y, z=screw_center_z)


def _flex_tip_rack_lid_feature_to_child_origin(
    child_labware: LabwareDefinition3,
) -> Point:
    """Returns the offset from a Flex tip rack lid locating feature to the child origin."""
    flex_tip_rack_lid_as_child = child_labware.features.get(
        "opentronsFlexTipRackLidAsChild"
    )
    assert flex_tip_rack_lid_as_child is not None

    return Point(x=0, y=0, z=flex_tip_rack_lid_as_child["matingZ"])


def slot_bottom_center_to_child_origin(
    child_labware: LabwareDefinition3,
) -> Point:
    """Returns offset from a parent slot's bottom center to the child origin."""
    slot_footprint_as_child = child_labware.features.get("slotFootprintAsChild")
    assert slot_footprint_as_child is not None

    x = (
        slot_footprint_as_child["frontRight"]["x"]
        + slot_footprint_as_child["backLeft"]["x"]
    ) / 2
    y = (
        slot_footprint_as_child["frontRight"]["y"]
        + slot_footprint_as_child["backLeft"]["y"]
    ) / 2
    z = slot_footprint_as_child["z"]

    return Point(x, y, z) * -1


def _slot_back_left_bottom_to_child_origin(
    child_labware: LabwareDefinition3,
) -> Point:
    """Returns offset from a parent slot's back left bottom to the child's origin."""
    slot_footprint_as_child = child_labware.features.get("slotFootprintAsChild")
    assert slot_footprint_as_child is not None

    x = slot_footprint_as_child["backLeft"]["x"]
    y = slot_footprint_as_child["backLeft"]["y"]
    z = slot_footprint_as_child["z"]

    return Point(x, y, z) * -1


def _child_back_left_bottom_position(child_labware: LabwareDefinition3) -> Point:
    """Get the back left bottom position from a v3 labware definition."""
    footprint_as_child = _get_labware_footprint_as_child(child_labware)

    return Point(
        x=footprint_as_child["backLeft"]["x"],
        y=footprint_as_child["frontRight"]["y"],
        z=footprint_as_child["z"],
    )


def _heater_shaker_universal_flat_adapter_feature_to_child_origin(
    child_labware: LabwareDefinition3,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point:
    """Returns the offset from a Heater Shaker Universal Flat Adapter locating feature to the child origin."""
    flat_well_support_as_child = child_labware.features.get(
        "flatSupportThermalCouplingAsChild"
    )

    assert flat_well_support_as_child is not None

    well_exterior_bottom_z = flat_well_support_as_child["wellExteriorBottomZ"]
    extents = child_labware.extents.total

    if underlying_ancestor_orientation == ModuleOrientation.LEFT:
        left_side_center_x = extents.backLeftBottom.x
        left_side_center_y = (extents.backLeftBottom.y + extents.frontRightTop.y) / 2

        return (
            Point(left_side_center_x, left_side_center_y, well_exterior_bottom_z) * -1
        )
    elif underlying_ancestor_orientation == ModuleOrientation.RIGHT:
        right_side_center_x = extents.frontRightTop.x
        right_side_center_y = (extents.backLeftBottom.y + extents.frontRightTop.y) / 2

        return (
            Point(right_side_center_x, right_side_center_y, well_exterior_bottom_z) * -1
        )

    else:
        raise InvalidLabwarePlacementError(
            feature_name="heaterShakerUniversalFlatAdapter",
            invalid_placement=ModuleOrientation.CENTER.value,
        )


def _screw_anchored_feature_to_child_origin(
    child_labware: LabwareDefinition3,
    underlying_ancestor_orientation: ModuleOrientation,
) -> Point:
    """Returns the offset from a `screwAnchoredAsChild` locating feature to the child origin."""
    screw_center = _get_screw_anchored_center_for_child(
        child_labware, underlying_ancestor_orientation
    )
    assert screw_center is not None

    screw_center_x = screw_center["x"]
    screw_center_y = screw_center["y"]
    screw_center_z = screw_center["z"]

    return Point(x=screw_center_x, y=screw_center_y, z=screw_center_z) * -1


def _get_child_labware_overlap_with_parent_labware(
    child_labware: LabwareDefinition, parent_labware_name: str
) -> Point:
    """Get the child labware's overlap with the parent labware's load name."""
    if isinstance(child_labware, LabwareDefinition3) and not hasattr(
        child_labware, "legacyStackingOffsetWithLabware"
    ):
        raise NotImplementedError(
            f"Labware {child_labware.metadata.displayName} contains no legacyStackingOffsetWithLabware. "
            f"Either add this property explictly on the definition or update your protocol's API Level."
        )

    overlap = (
        child_labware.stackingOffsetWithLabware.get(parent_labware_name)
        if isinstance(child_labware, LabwareDefinition2)
        else child_labware.legacyStackingOffsetWithLabware.get(parent_labware_name)
    )

    if overlap is None:
        overlap = (
            child_labware.stackingOffsetWithLabware.get("default")
            if isinstance(child_labware, LabwareDefinition2)
            else child_labware.legacyStackingOffsetWithLabware.get("default")
        )

    if overlap is None:
        if isinstance(child_labware, LabwareDefinition3):
            raise ValueError(
                f"No default labware overlap specified for parent labware: {parent_labware_name} "
                f"in legacyStackingOffsetWithLabware."
            )
        else:
            raise ValueError(
                f"No default labware overlap specified for parent labware: {parent_labware_name} "
                f"in stackingOffsetWithLabware."
            )
    else:
        return Point.from_xyz_attrs(overlap)


def _get_child_labware_overlap_with_parent_module(
    child_labware: LabwareDefinition,
    parent_module_model: ModuleModel,
    deck_definition: DeckDefinitionV5,
) -> Point:
    """Get the child labware's overlap with the parent module model."""
    child_labware_overlap = child_labware.stackingOffsetWithModule.get(
        str(parent_module_model.value)
    )
    if not child_labware_overlap:
        if _is_thermocycler_on_ot2(parent_module_model, deck_definition):
            return _OFFSET_ON_TC_OT2
        else:
            return Point(x=0, y=0, z=0)

    return Point.from_xyz_attrs(child_labware_overlap)


def _feature_exception_offsets(
    parent_deck_item: _LabwareStackupDefinition,
    deck_definition: DeckDefinitionV5,
) -> Point:
    """These offsets are intended for legacy reasons only and should generally be avoided post labware schema 2.

    If you need to make exceptions for a parent-child stackup, use the `custom` locating feature.
    """
    if isinstance(parent_deck_item, ModuleDefinition) and _is_thermocycler_on_ot2(
        parent_deck_item.model, deck_definition
    ):
        return _OFFSET_ON_TC_OT2
    else:
        return Point(x=0, y=0, z=0)


def _is_thermocycler_on_ot2(
    parent_module_model: ModuleModel,
    deck_definition: DeckDefinitionV5,
) -> bool:
    """Whether the given parent module is a thermocycler with the current deck being an OT2 deck."""
    robot_model = deck_definition["robot"]["model"]
    return (
        parent_module_model
        in [ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2]
        and robot_model == "OT-2 Standard"
    )


def _get_labware_footprint_as_child(
    labware: LabwareDefinition3,
) -> SlotFootprintAsChildFeature:
    """Get the SlotFootprintAsChildFeature for labware definitions."""
    footprint_as_child = labware.features.get("slotFootprintAsChild")
    if footprint_as_child is None:
        raise MissingLocatingFeatureError(
            labware_name=labware.metadata.displayName,
            required_feature="slotFootprintAsChild",
        )
    else:
        return footprint_as_child


def _total_nominal_gripper_offsets(
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
    slot_name: DeckSlotName,
    deck_definition: DeckDefinitionV5,
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
) -> _GripperOffsets:
    """Get the total of the offsets to be used to pick up and drop labware."""
    top_most_lw_definition, top_most_lw_location = stackup_lw_info_top_to_bottom[0]
    special_offsets = _get_special_gripper_offsets(
        stackup_lw_info_top_to_bottom, underlying_ancestor_definition
    )

    if isinstance(
        top_most_lw_location,
        (ModuleLocation, DeckSlotLocation, AddressableAreaLocation),
    ):
        offsets = _nominal_gripper_offsets_for_location(
            labware_location=top_most_lw_location,
            labware_definition=top_most_lw_definition,
            slot_name=slot_name,
            deck_definition=deck_definition,
            underlying_ancestor_definition=underlying_ancestor_definition,
        )

        pick_up_offset = Point.from_xyz_attrs(offsets.pickUpOffset)
        drop_offset = Point.from_xyz_attrs(offsets.dropOffset)

        return _GripperOffsets(
            pick_up_offset=pick_up_offset + special_offsets.pick_up_offset,
            drop_offset=drop_offset + special_offsets.drop_offset,
        )
    else:
        # If it's a labware on a labware (most likely an adapter),
        # we calculate the offset as sum of offsets for the direct parent labware
        # and the underlying non-labware parent location.
        direct_parent_def, direct_parent_loc = stackup_lw_info_top_to_bottom[1]
        direct_parent_offsets = _nominal_gripper_offsets_for_location(
            labware_location=direct_parent_loc,
            labware_definition=direct_parent_def,
            slot_name=slot_name,
            deck_definition=deck_definition,
            underlying_ancestor_definition=underlying_ancestor_definition,
        )

        top_most_offsets = _nominal_gripper_offsets_for_location(
            labware_location=top_most_lw_location,
            labware_definition=top_most_lw_definition,
            slot_name=slot_name,
            deck_definition=deck_definition,
            underlying_ancestor_definition=underlying_ancestor_definition,
        )

        pick_up_offset = Point.from_xyz_attrs(
            direct_parent_offsets.pickUpOffset
        ) + Point.from_xyz_attrs(top_most_offsets.pickUpOffset)
        drop_offset = Point.from_xyz_attrs(
            direct_parent_offsets.dropOffset
        ) + Point.from_xyz_attrs(top_most_offsets.dropOffset)

        return _GripperOffsets(
            pick_up_offset=pick_up_offset + special_offsets.pick_up_offset,
            drop_offset=drop_offset + special_offsets.drop_offset,
        )


# TODO(jh, 08-15-25): Return _GripperOffsets instead of LabwareMovementOffsetData.
def _nominal_gripper_offsets_for_location(
    labware_location: LabwareLocation,
    labware_definition: LabwareDefinition,
    slot_name: DeckSlotName,
    deck_definition: DeckDefinitionV5,
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
) -> LabwareMovementOffsetData:
    """Provide the default gripper offset data for the given location type."""
    if isinstance(labware_location, (DeckSlotLocation, AddressableAreaLocation)):
        offsets = _get_deck_default_gripper_offsets(deck_definition)
    elif isinstance(labware_location, ModuleLocation):
        offsets = _get_module_default_gripper_offsets(underlying_ancestor_definition)  # type: ignore[arg-type]
    else:
        offsets = _labware_gripper_offsets(
            top_most_lw_definition=labware_definition, slot_name=slot_name
        )
    return offsets or LabwareMovementOffsetData(
        pickUpOffset=LabwareOffsetVector(x=0, y=0, z=0),
        dropOffset=LabwareOffsetVector(x=0, y=0, z=0),
    )


def _get_deck_default_gripper_offsets(
    deck_definition: DeckDefinitionV5,
) -> Optional[LabwareMovementOffsetData]:
    """Get the deck's default gripper offsets."""
    parsed_offsets = deck_definition.get("gripperOffsets", {}).get("default")
    return (
        LabwareMovementOffsetData(
            pickUpOffset=LabwareOffsetVector(
                x=parsed_offsets["pickUpOffset"]["x"],
                y=parsed_offsets["pickUpOffset"]["y"],
                z=parsed_offsets["pickUpOffset"]["z"],
            ),
            dropOffset=LabwareOffsetVector(
                x=parsed_offsets["dropOffset"]["x"],
                y=parsed_offsets["dropOffset"]["y"],
                z=parsed_offsets["dropOffset"]["z"],
            ),
        )
        if parsed_offsets
        else None
    )


def _get_module_default_gripper_offsets(
    module_definition: ModuleDefinition,
) -> Optional[LabwareMovementOffsetData]:
    """Get the deck's default gripper offsets."""
    offsets = module_definition.gripperOffsets
    return offsets.get("default") if offsets else None


def _labware_gripper_offsets(
    top_most_lw_definition: LabwareDefinition, slot_name: DeckSlotName
) -> Optional[LabwareMovementOffsetData]:
    """Provide the most appropriate gripper offset data for the specified labware.

    We check the types of gripper offsets available for the labware ("default" or slot-based)
    and return the most appropriate one for the overall location of the labware.
    Currently, only module adapters (specifically, the H/S universal flat adapter)
    have non-default offsets that are specific to location of the module on deck,
    so, this code only checks for the presence of those known offsets.
    """
    slot_based_offset = _get_child_gripper_offsets(
        top_most_lw_definition=top_most_lw_definition, slot_name=slot_name
    )
    return slot_based_offset or _get_child_gripper_offsets(
        top_most_lw_definition=top_most_lw_definition, slot_name=None
    )


def _get_child_gripper_offsets(
    top_most_lw_definition: LabwareDefinition,
    slot_name: DeckSlotName | None,
) -> Optional[LabwareMovementOffsetData]:
    """Get the grip offsets that a labware says should be applied to children stacked atop it.

    If `slot_name` is provided, returns the gripper offsets that the parent labware definition
    specifies just for that slot, or `None` if the labware definition doesn't have an
    exact match.

    If `slot_name` is `None`, returns the gripper offsets that the parent labware
    definition designates as "default," or `None` if it doesn't designate any as such.
    """
    parsed_offsets = top_most_lw_definition.gripperOffsets
    offset_key = slot_name.id if slot_name else "default"

    if parsed_offsets is None or offset_key not in parsed_offsets:
        return None
    else:
        return LabwareMovementOffsetData(
            pickUpOffset=LabwareOffsetVector.model_construct(
                x=parsed_offsets[offset_key].pickUpOffset.x,
                y=parsed_offsets[offset_key].pickUpOffset.y,
                z=parsed_offsets[offset_key].pickUpOffset.z,
            ),
            dropOffset=LabwareOffsetVector.model_construct(
                x=parsed_offsets[offset_key].dropOffset.x,
                y=parsed_offsets[offset_key].dropOffset.y,
                z=parsed_offsets[offset_key].dropOffset.z,
            ),
        )


def _get_special_gripper_offsets(
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
) -> _GripperOffsets:
    """Handles all special-cased gripper offsets."""
    tc_lid_offsets = (
        _get_tc_lid_gripper_offsets(
            stackup_lw_info_top_to_bottom, underlying_ancestor_definition
        )
    ) or _GripperOffsets(drop_offset=Point(), pick_up_offset=Point())

    ar_lid_offsets = (
        _get_absorbance_reader_lid_gripper_offsets(stackup_lw_info_top_to_bottom)
    ) or _GripperOffsets(drop_offset=Point(), pick_up_offset=Point())

    return _GripperOffsets(
        pick_up_offset=tc_lid_offsets.pick_up_offset + ar_lid_offsets.pick_up_offset,
        drop_offset=tc_lid_offsets.pick_up_offset + ar_lid_offsets.drop_offset,
    )


def _get_tc_lid_gripper_offsets(
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
) -> _GripperOffsets | None:
    top_most_lw_def, top_most_lw_loc = stackup_lw_info_top_to_bottom[0]

    if isinstance(top_most_lw_loc, OnLabwareLocation):
        bottom_most_lw_location = stackup_lw_info_top_to_bottom[-1][1]

        # This is done as a workaround for some TC geometry inaccuracies.
        # See PLAT-579 for context.
        if (
            isinstance(bottom_most_lw_location, ModuleLocation)
            and getattr(underlying_ancestor_definition, "model", None)
            == ModuleModel.THERMOCYCLER_MODULE_V2
            and validate_definition_is_lid(top_most_lw_def)
        ):
            # It is intentional to use the `pickUpOffset` in both the gripper pick up and drop cases.
            if "lidOffsets" in top_most_lw_def.gripperOffsets.keys():
                offset = Point(
                    x=top_most_lw_def.gripperOffsets["lidOffsets"].pickUpOffset.x,
                    y=top_most_lw_def.gripperOffsets["lidOffsets"].pickUpOffset.y,
                    z=top_most_lw_def.gripperOffsets["lidOffsets"].pickUpOffset.z,
                )
                return _GripperOffsets(pick_up_offset=offset, drop_offset=offset)
            else:
                raise LabwareOffsetDoesNotExistError(
                    f"Labware Definition {top_most_lw_def.parameters.loadName} does not contain required field 'lidOffsets' of 'gripperOffsets'."
                )

    return None


def _get_absorbance_reader_lid_gripper_offsets(
    stackup_lw_info_top_to_bottom: list[tuple[LabwareDefinition, LabwareLocation]],
) -> _GripperOffsets | None:
    top_most_lw_definition = stackup_lw_info_top_to_bottom[0][0]
    load_name = top_most_lw_definition.parameters.loadName

    if is_absorbance_reader_lid(load_name):
        # todo(mm, 2024-11-06): This is only correct in the special case of an
        # absorbance reader lid. Its definition currently puts the offsets for *itself*
        # in the property that's normally meant for offsets for its *children.*
        offsets = _get_child_gripper_offsets(top_most_lw_definition, slot_name=None)

        if offsets is None:
            raise ValueError(
                "Expected gripper offsets for absorbance reader lid to be defined."
            )
        else:
            return _GripperOffsets(
                pick_up_offset=Point.from_xyz_attrs(offsets.pickUpOffset),
                drop_offset=Point.from_xyz_attrs(offsets.dropOffset),
            )

    else:
        return None



"""Test Flex Stacker close latch command implementation."""

import pytest
from decoy import Decoy

from opentrons.hardware_control.modules import FlexStacker
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.unsafe.unsafe_stacker_close_latch import (
    UnsafeFlexStackerCloseLatchImpl,
    UnsafeFlexStackerCloseLatchParams,
    UnsafeFlexStackerCloseLatchResult,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView


@pytest.fixture
def subject(
    state_view: StateView, equipment: EquipmentHandler, model_utils: ModelUtils
) -> UnsafeFlexStackerCloseLatchImpl:
    """Get a UnsafeFlexStackerCloseLatch command to test."""
    return UnsafeFlexStackerCloseLatchImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


async def test_close_latch_command(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerCloseLatchImpl,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should return a success data."""
    data = UnsafeFlexStackerCloseLatchParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=0,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    result = await subject.execute(data)

    decoy.verify(await stacker_hardware.close_latch(), times=1)

    assert result == SuccessData(public=UnsafeFlexStackerCloseLatchResult())



from unittest.mock import MagicMock

import pytest

from opentrons import hardware_control as hc
from opentrons.drivers import rpi_drivers
from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.hardware_control.types import BoardRevision


async def test_gpio_setup() -> None:
    # Test without DTOVERLAY path
    # Board revision should be defaulted to 2.1
    backend = await hc.Controller.build(config=None)
    assert str(backend.board_revision) == "2.1"


@pytest.mark.parametrize(
    "revision,a_current",
    [
        (hc.types.BoardRevision.OG, 0.8),
        (hc.types.BoardRevision.A, 0.5),
        (hc.types.BoardRevision.B, 0.5),
        (hc.types.BoardRevision.C, 0.5),
    ],
)
async def test_current_setting(
    monkeypatch: pytest.MonkeyPatch, revision: BoardRevision, a_current: float
) -> None:
    mock_gpio = MagicMock(spec=SimulatingGPIOCharDev)

    async def fake_side_effect() -> None:
        pass

    mock_gpio.setup.side_effect = fake_side_effect
    mock_gpio.board_rev = revision
    mock_build = MagicMock(spec=rpi_drivers, return_value=mock_gpio)
    monkeypatch.setattr(hc.backends.controller, "build_gpio_chardev", mock_build)
    backend = await hc.Controller.build(config=None)
    assert backend.board_revision == revision
    assert backend._smoothie_driver._active_current_settings.now["A"] == a_current



from datetime import datetime

import pytest

from opentrons.calibration_storage.types import (
    CalibrationStatus,
    SourceType,
)
from opentrons.hardware_control import API, ThreadManager
from opentrons.hardware_control.robot_calibration import (
    DeckCalibration,
    RobotCalibration,
)
from opentrons.hardware_control.util import DeckTransformState


@pytest.mark.ot2_only  # ot3 attitude is always correct
async def test_validating_attitude(hardware: ThreadManager[API]) -> None:
    inrange_matrix = [[1.0, 0, 1], [0.0, 1, 2], [0.0, 0, 1]]
    deck_cal = DeckCalibration(
        attitude=inrange_matrix,
        last_modified=datetime.now(),
        source=SourceType.user,
        status=CalibrationStatus(),
    )

    hardware.set_robot_calibration(RobotCalibration(deck_calibration=deck_cal))

    assert hardware.validate_calibration() == DeckTransformState.OK

    identity_matrix = [[1.0, 0, 0], [0.0, 1, 0], [0.0, 0, 1]]
    deck_cal.attitude = identity_matrix
    deck_cal.last_modified = None
    hardware.set_robot_calibration(RobotCalibration(deck_calibration=deck_cal))
    assert hardware.validate_calibration() == DeckTransformState.IDENTITY

    singular_matrix = [[0.0, 0, 0], [0.0, 0, 0], [0.0, 0, 1]]
    deck_cal.attitude = singular_matrix
    hardware.set_robot_calibration(RobotCalibration(deck_calibration=deck_cal))

    assert hardware.validate_calibration() == DeckTransformState.SINGULARITY



from typing import AsyncGenerator

import pytest

from .build_module import build_module
from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.modules import TempDeck


@pytest.fixture
async def tempdeck(
    emulator_settings: Settings,
    emulation_app: None,
    execution_manager: ExecutionManager,
    poll_interval_seconds: float,
) -> AsyncGenerator[TempDeck, None]:
    module = await build_module(
        TempDeck,
        port=emulator_settings.temperature_proxy.driver_port,
        execution_manager=execution_manager,
        poll_interval_seconds=poll_interval_seconds,
    )
    try:
        yield module
    finally:
        await module.cleanup()


def test_device_info(tempdeck: TempDeck) -> None:
    """It should have the device info."""
    assert {
        "model": "temp_deck_v20",
        "serial": "temperature_emulator",
        "version": "v2.0.1",
    } == tempdeck.device_info


async def test_set_temperature(tempdeck: TempDeck) -> None:
    """It should set the temperature and return when target is reached."""
    assert tempdeck.live_data == {
        "status": "idle",
        "data": {"currentTemp": 0, "targetTemp": None},
    }

    await tempdeck.start_set_temperature(10)
    await tempdeck.await_temperature(None)

    assert tempdeck.live_data == {
        "status": "holding at target",
        "data": {"currentTemp": 10, "targetTemp": 10},
    }


async def test_start_set_temperature_cool(tempdeck: TempDeck) -> None:
    """It should set the temperature and return and wait for temperature."""
    current = tempdeck.temperature
    new_temp = current - 20

    await tempdeck.start_set_temperature(new_temp)
    assert tempdeck.live_data == {
        "status": "cooling",
        "data": {"currentTemp": pytest.approx(current, abs=5), "targetTemp": new_temp},
    }

    # Wait for temperature to be reached
    await tempdeck.await_temperature(awaiting_temperature=new_temp)
    assert tempdeck.live_data == {
        "status": "holding at target",
        "data": {"currentTemp": new_temp, "targetTemp": new_temp},
    }


async def test_start_set_temperature_heat(tempdeck: TempDeck) -> None:
    """It should set the temperature and return and wait for temperature."""
    current = tempdeck.temperature
    new_temp = current + 20

    await tempdeck.start_set_temperature(new_temp)
    assert tempdeck.live_data == {
        "status": "heating",
        "data": {"currentTemp": pytest.approx(current, abs=5), "targetTemp": new_temp},
    }

    # Wait for temperature to be reached
    await tempdeck.await_temperature(awaiting_temperature=new_temp)
    assert tempdeck.live_data == {
        "status": "holding at target",
        "data": {"currentTemp": new_temp, "targetTemp": new_temp},
    }


async def test_deactivate(tempdeck: TempDeck) -> None:
    """It should deactivate and move to room temperature"""
    await tempdeck.deactivate()
    await tempdeck.await_temperature(awaiting_temperature=23)

    assert tempdeck.live_data == {
        "status": "idle",
        "data": {"currentTemp": 23, "targetTemp": None},
    }


async def test_forcible_deactivate(
    tempdeck: TempDeck, execution_manager: ExecutionManager
) -> None:
    """Can override wait_for_is_running."""
    await execution_manager.pause()
    await tempdeck.deactivate(must_be_running=False)



"""Test robot.move-axes-to commands."""

from decoy import Decoy

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.protocols.types import FlexRobotType
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.robot.move_axes_to import (
    MoveAxesToImplementation,
    MoveAxesToParams,
    MoveAxesToResult,
)
from opentrons.protocol_engine.execution import GantryMover
from opentrons.protocol_engine.types import MotorAxis


async def test_move_axes_to_implementation(
    decoy: Decoy,
    gantry_mover: GantryMover,
    ot3_hardware_api: HardwareControlAPI,
) -> None:
    """Test the `robot.moveAxesTo` implementation.

    It should call `MovementHandler.move_mount_to` with the
    correct coordinates.
    """
    subject = MoveAxesToImplementation(
        gantry_mover=gantry_mover,
        hardware_api=ot3_hardware_api,
    )

    params = MoveAxesToParams(
        axis_map={MotorAxis.X: 10, MotorAxis.Y: 10, MotorAxis.EXTENSION_Z: 20},
        critical_point={MotorAxis.X: 1, MotorAxis.Y: 1, MotorAxis.EXTENSION_Z: 0},
        speed=567.8,
    )

    # Flex shape
    decoy.when(ot3_hardware_api.get_robot_type()).then_return(FlexRobotType)
    decoy.when(
        await gantry_mover.move_axes(
            axis_map=params.axis_map,
            speed=params.speed,
            critical_point=params.critical_point,
        )
    ).then_return({MotorAxis.X: 10, MotorAxis.Y: 10, MotorAxis.EXTENSION_Z: 20})
    result = await subject.execute(params=params)

    assert result == SuccessData(
        public=MoveAxesToResult(
            position={MotorAxis.X: 10, MotorAxis.Y: 10, MotorAxis.EXTENSION_Z: 20}
        )
    )



"""Tests for performance_helpers."""

from pathlib import Path

from opentrons.util.performance_helpers import (
    _get_robot_activity_tracker,
    _StubbedTracker,
)


def test_return_function_unchanged() -> None:
    """Test that the function is returned unchanged when using _StubbedTracker."""
    tracker = _StubbedTracker(Path("/path/to/storage"), True)

    def func_to_track() -> None:
        pass

    assert tracker.track("ANALYZING_PROTOCOL")(func_to_track) is func_to_track


def test_singleton_tracker() -> None:
    """Test that the tracker is a singleton."""
    tracker = _get_robot_activity_tracker()
    tracker2 = _get_robot_activity_tracker()
    assert tracker is tracker2



"""Test Thermocycler set block temperature command implementation."""

from decoy import Decoy, matchers

from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_engine.actions import Action, ActionDispatcher, StartTaskAction
from opentrons.protocol_engine.commands import thermocycler as tc_commands
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.thermocycler.set_target_block_temperature import (
    SetTargetBlockTemperatureImpl,
    SetTargetBlockTemperatureResult,
)
from opentrons.protocol_engine.execution import EquipmentHandler, TaskHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types.tasks import Task


async def test_set_target_block_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
) -> None:
    """It should be able to set the specified module's target temperature."""
    subject = SetTargetBlockTemperatureImpl(
        state_view=state_view, equipment=equipment, task_handler=real_task_handler
    )

    data = tc_commands.SetTargetBlockTemperatureParams(
        moduleId="input-thermocycler-id",
        celsius=12.3,
        blockMaxVolumeUl=50.2,
        holdTimeSeconds=123456,
        taskId="taskId",
    )
    tc_commands.SetTargetBlockTemperatureResult(
        targetBlockTemperature=45.6, taskId="taskId"
    )

    tc_module_substate = decoy.mock(cls=ThermocyclerModuleSubState)
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(
        state_view.modules.get_thermocycler_module_substate("input-thermocycler-id")
    ).then_return(tc_module_substate)
    decoy.when(tc_module_substate.module_id).then_return(
        ThermocyclerModuleId("thermocycler-id")
    )
    decoy.when(model_utils.ensure_id("taskId")).then_return("taskId")

    # Stub temperature validation from TC module view
    decoy.when(tc_module_substate.validate_target_block_temperature(12.3)).then_return(
        45.6
    )

    # Stub volume validation from TC module view
    decoy.when(tc_module_substate.validate_max_block_volume(50.2)).then_return(77.6)
    # Stub hold time validation from TC module view
    decoy.when(tc_module_substate.validate_hold_time(123456)).then_return(654321)

    # Get attached hardware modules
    decoy.when(
        equipment.get_module_hardware_api(ThermocyclerModuleId("thermocycler-id"))
    ).then_return(tc_hardware)

    task: Task | None = None

    def _capture_task(action: Action) -> None:
        nonlocal task
        assert isinstance(action, StartTaskAction)
        task = action.task

    decoy.when(
        action_dispatcher.dispatch(StartTaskAction(task=matchers.Anything()))  # type: ignore[func-returns-value]
    ).then_do(_capture_task)

    result = await subject.execute(data)
    assert task is not None
    await task.asyncioTask

    decoy.verify(
        await tc_hardware.set_target_block_temperature(
            celsius=45.6,
            hold_time_seconds=654321,
            volume=77.6,
            ramp_rate=None,
        ),
    )
    decoy.verify(await tc_hardware.wait_for_block_target())
    assert result == SuccessData(
        public=SetTargetBlockTemperatureResult(
            targetBlockTemperature=45.6, taskId="taskId"
        )
    )



import asyncio

import pytest

from opentrons_shared_data.errors.exceptions import ExecutionCancelledError

from opentrons.hardware_control import (
    ExecutionManager,
    ExecutionState,
)


async def test_state_machine() -> None:
    """
    Test that an execution manager's state is RUNNING on init
    and PAUSE when it when pause is called, unless CANCELLED
    """
    exec_mgr = ExecutionManager()
    assert await exec_mgr.get_state() == ExecutionState.RUNNING

    # passes through on wait_for_is_running if state is RUNNING
    await asyncio.wait_for(exec_mgr.wait_for_is_running(), timeout=0.2)

    await exec_mgr.pause()
    assert await exec_mgr.get_state() == ExecutionState.PAUSED

    with pytest.raises(asyncio.TimeoutError):
        # should stall on wait_for_is_running when state is PAUSED
        await asyncio.wait_for(exec_mgr.wait_for_is_running(), timeout=0.2)

    await exec_mgr.resume()
    assert await exec_mgr.get_state() == ExecutionState.RUNNING

    await exec_mgr.cancel()
    assert await exec_mgr.get_state() == ExecutionState.CANCELLED

    with pytest.raises(ExecutionCancelledError):
        # attempting to pause when CANCELLED should raise
        await exec_mgr.pause()

    with pytest.raises(ExecutionCancelledError):
        # should raise on wait_for_is_running when state is CANCELLED
        await asyncio.wait_for(exec_mgr.wait_for_is_running(), timeout=0.2)

    await exec_mgr.reset()
    assert await exec_mgr.get_state() == ExecutionState.RUNNING


async def test_cancel_tasks() -> None:
    """
    Test that an execution manager cancels all un-protected
    running asyncio Tasks when cancel is called
    """

    async def fake_task() -> None:
        while True:
            await asyncio.sleep(1)

    exec_mgr = ExecutionManager()

    loop = asyncio.get_running_loop()

    cancellable_task = loop.create_task(fake_task())
    exec_mgr.register_cancellable_task(cancellable_task)

    other_task = loop.create_task(fake_task())

    # current, cancellable, and other
    assert len(asyncio.all_tasks(loop)) == 3
    assert len([t for t in asyncio.all_tasks(loop) if t.cancelled()]) == 0

    await exec_mgr.cancel()
    await asyncio.sleep(0.1)

    all_tasks = asyncio.all_tasks(loop)
    assert len(all_tasks) == 2  # current and other
    assert other_task in all_tasks
    assert cancellable_task not in all_tasks

    other_task.cancel()



from datetime import datetime
from typing import Callable, Optional
from unittest import mock

import pytest

from opentrons_shared_data.errors.exceptions import MotionFailedError
from opentrons_shared_data.gripper import GripperModel

from opentrons.calibration_storage import types as cal_types
from opentrons.config import gripper_config
from opentrons.hardware_control.instruments.ot3 import gripper, instrument_calibration
from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    GripperCalibrationOffset,
    GripperJawWidthData,
)
from opentrons.hardware_control.types import CriticalPoint
from opentrons.types import Point

fake_gripper_conf = gripper_config.load(GripperModel.v1)


@pytest.fixture
def fake_offset() -> GripperCalibrationOffset:
    return instrument_calibration.load_gripper_calibration_offset("fakeid123")


@pytest.fixture
def fake_jaw_cal(request: pytest.FixtureRequest) -> GripperJawWidthData:
    request.node.add_marker("ot3_only")
    return instrument_calibration.load_gripper_jaw_width("fakeid123")


@pytest.mark.ot3_only
def test_id_get_added_to_dict(fake_offset: GripperCalibrationOffset) -> None:
    gripr = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")
    assert gripr.as_dict()["gripper_id"] == "fakeid123"


@pytest.mark.xfail
@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "override,result_accessor",
    [
        (None, lambda g: g._jaw_center_offset),
        (CriticalPoint.GRIPPER_JAW_CENTER, lambda g: g._jaw_center_offset),
        (
            CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN,
            lambda g: g._front_calibration_pin_offset,
        ),
        (
            CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN,
            lambda g: g._rear_calibration_pin_offset,
        ),
    ],
)
def test_critical_point(
    override: Optional[CriticalPoint],
    result_accessor: Callable[[gripper.Gripper], Point],
    fake_offset: GripperCalibrationOffset,
) -> None:
    gripr = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")
    assert gripr.critical_point(override) == result_accessor(gripr)


@pytest.mark.ot3_only
def test_load_gripper_cal_offset(fake_offset: GripperCalibrationOffset) -> None:
    gripr = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")
    # if offset data do not exist, loaded values should match DEFAULT
    assert gripr._calibration_offset.offset == Point(
        *gripper_config.DEFAULT_GRIPPER_CALIBRATION_OFFSET
    )


@pytest.mark.ot3_only
def test_gripper_default_jaw_width_calibration(
    fake_jaw_cal: GripperJawWidthData,
    fake_offset: GripperCalibrationOffset,
) -> None:
    gripr = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")
    assert gripr._jaw_max_offset is None
    assert gripr._encoder_position_at_jaw_closed is None


@pytest.mark.parametrize("loaded_encoder_pos", [24.4, None])
@pytest.mark.parametrize("existing_encoder_pos", [24.4, None])
@pytest.mark.ot3_only
def test_gripper_has_jaw_width_calibration(
    fake_jaw_cal: GripperJawWidthData,
    fake_offset: GripperCalibrationOffset,
    loaded_encoder_pos: float,
    existing_encoder_pos: float,
) -> None:
    gripr = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")
    gripr._encoder_position_at_jaw_closed = existing_encoder_pos
    with mock.patch(
        "opentrons.hardware_control.instruments.ot3.gripper.load_gripper_jaw_width",
        return_value=GripperJawWidthData(
            source=fake_jaw_cal.source,
            status=fake_jaw_cal.status,
            encoder_position_at_jaw_closed=loaded_encoder_pos,
            last_modified=datetime.now(),
        ),
        autospec=True,
    ) as fake_load_jaw_width:
        has_cal = gripr.has_jaw_width_calibration
        mock_save_gripper_jaw_width_data = mock.Mock()
        mock.patch(
            "opentrons.hardware_control.instruments.ot3.gripper.save_gripper_jaw_width_data",
            return_value=mock_save_gripper_jaw_width_data(),
        )
        # if gripper._encoder_position_at_jaw_closed has no value:
        if existing_encoder_pos is None:
            # gripper should try to load jaw width from the robot fs
            fake_load_jaw_width.assert_called_once()
            if loaded_encoder_pos is None:
                assert has_cal is False
            else:
                # if robot fs has gripper jaw width data,
                # it should get saved to the gripper object and has_cal return true
                mock_save_gripper_jaw_width_data.assert_called_once()
                assert has_cal is True
        # if gripper._encoder_position_at_jaw_closed has a value:
        else:
            # gripper doesn't try to load from the robot and returns true
            fake_load_jaw_width.assert_not_called()
            assert has_cal is True
            mock_save_gripper_jaw_width_data.assert_called_once()


@pytest.mark.ot3_only
def test_reload_instrument_cal_ot3(fake_offset: GripperCalibrationOffset) -> None:
    old_gripper = gripper.Gripper(
        fake_gripper_conf,
        fake_offset,
        "fakeid123",
    )
    # if only calibration is changed
    new_cal = instrument_calibration.GripperCalibrationOffset(
        offset=Point(3, 4, 5),
        source=cal_types.SourceType.user,
        status=cal_types.CalibrationStatus(),
    )
    new_gripper, skip = gripper._reload_gripper(
        old_gripper.config, old_gripper, new_cal
    )

    # it's the same gripper
    assert new_gripper == old_gripper
    # jaw offset should persists as well
    assert new_gripper._jaw_max_offset == old_gripper._jaw_max_offset
    # we said upstream could skip
    assert skip


@pytest.mark.ot3_only
def test_reload_instrument_cal_ot3_conf_changed(
    fake_offset: GripperCalibrationOffset,
) -> None:
    old_gripper = gripper.Gripper(
        fake_gripper_conf,
        fake_offset,
        "fakeid123",
    )
    new_conf = fake_gripper_conf.model_copy(
        update={
            "grip_force_profile": fake_gripper_conf.grip_force_profile.model_copy(
                update={"default_grip_force": 1}
            )
        },
        deep=True,
    )
    assert new_conf != old_gripper.config

    new_gripper, skip = gripper._reload_gripper(new_conf, old_gripper, fake_offset)

    # it's not the same gripper
    assert new_gripper != old_gripper
    # do not pass in the old jaw max offse
    assert not new_gripper._jaw_max_offset
    # we said upstream could skip
    assert not skip


@pytest.mark.ot3_only
def test_jaw_calibration_error_checking() -> None:
    subject = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")  # type: ignore[arg-type]
    with pytest.raises(MotionFailedError):
        subject.update_jaw_open_position_from_closed_position(0)


@pytest.mark.ot3_only
def test_jaw_calibration() -> None:
    subject = gripper.Gripper(fake_gripper_conf, fake_offset, "fakeid123")  # type: ignore[arg-type]
    subject.update_jaw_open_position_from_closed_position(
        (
            fake_gripper_conf.geometry.jaw_width["max"]
            - fake_gripper_conf.geometry.jaw_width["min"]
            + 2
        )
        / 2
    )
    assert subject.max_jaw_width == fake_gripper_conf.geometry.jaw_width["max"] + 2



"""Tests for the PythonAndLegacyRunner, JsonRunner & LiveRunner classes."""

from datetime import datetime
from pathlib import Path
from typing import List, Type, Union, cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2
from opentrons_shared_data.labware.types import (
    LabwareDefinition as LabwareDefinitionTypedDict,
)
from opentrons_shared_data.protocol.models import ProtocolSchemaV6, ProtocolSchemaV7
from opentrons_shared_data.protocol.types import (
    JsonProtocol as LegacyJsonProtocolDict,
)

from opentrons import protocol_reader
from opentrons.hardware_control import API as HardwareAPI
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import ProtocolContext
from opentrons.protocol_engine import (
    CommandStatus,
    Liquid,
    ProtocolEngine,
)
from opentrons.protocol_engine import (
    commands as pe_commands,
)
from opentrons.protocol_engine import (
    errors as pe_errors,
)
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.types import PostRunHardwareState
from opentrons.protocol_reader import (
    JsonProtocolConfig,
    ProtocolSource,
    PythonProtocolConfig,
)
from opentrons.protocol_runner import (
    AnyRunner,
    JsonRunner,
    LiveRunner,
    PythonAndLegacyRunner,
    create_protocol_runner,
)
from opentrons.protocol_runner.json_file_reader import JsonFileReader
from opentrons.protocol_runner.json_translator import JsonTranslator
from opentrons.protocol_runner.legacy_context_plugin import LegacyContextPlugin
from opentrons.protocol_runner.python_protocol_wrappers import (
    ProtocolContextCreator,
    PythonAndLegacyFileReader,
    PythonProtocolExecutor,
)
from opentrons.protocol_runner.task_queue import TaskQueue
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parse import PythonParseMode
from opentrons.protocols.types import JsonProtocol, PythonProtocol
from opentrons.util.broker import Broker


@pytest.fixture
def protocol_engine(decoy: Decoy) -> ProtocolEngine:
    """Get a mocked out ProtocolEngine dependency."""
    return decoy.mock(cls=ProtocolEngine)


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mocked out HardwareAPI dependency."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def task_queue(decoy: Decoy) -> TaskQueue:
    """Get a mocked out TaskQueue dependency."""
    return decoy.mock(cls=TaskQueue)


@pytest.fixture
def json_file_reader(decoy: Decoy) -> JsonFileReader:
    """Get a mocked out JsonFileReader dependency."""
    return decoy.mock(cls=JsonFileReader)


@pytest.fixture
def json_translator(decoy: Decoy) -> JsonTranslator:
    """Get a mocked out JsonTranslator dependency."""
    return decoy.mock(cls=JsonTranslator)


@pytest.fixture
def python_and_legacy_file_reader(decoy: Decoy) -> PythonAndLegacyFileReader:
    """Get a mocked out PythonAndLegacyFileReader dependency."""
    return decoy.mock(cls=PythonAndLegacyFileReader)


@pytest.fixture
def protocol_context_creator(decoy: Decoy) -> ProtocolContextCreator:
    """Get a mocked out ProtocolContextCreator dependency."""
    return decoy.mock(cls=ProtocolContextCreator)


@pytest.fixture
def python_protocol_executor(decoy: Decoy) -> PythonProtocolExecutor:
    """Get a mocked out PythonProtocolExecutor dependency."""
    return decoy.mock(cls=PythonProtocolExecutor)


@pytest.fixture(autouse=True)
def use_mock_extract_labware_definitions(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Replace protocol_reader.extract_labware_definitions() with a Decoy mock."""
    monkeypatch.setattr(
        protocol_reader,
        "extract_labware_definitions",
        decoy.mock(func=protocol_reader.extract_labware_definitions),
    )


@pytest.fixture
def json_runner_subject(
    protocol_engine: ProtocolEngine,
    hardware_api: HardwareAPI,
    task_queue: TaskQueue,
    json_file_reader: JsonFileReader,
    json_translator: JsonTranslator,
) -> JsonRunner:
    """Get a JsonRunner test subject with mocked dependencies."""
    return JsonRunner(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        task_queue=task_queue,
        json_file_reader=json_file_reader,
        json_translator=json_translator,
    )


@pytest.fixture
def python_runner_subject(
    protocol_engine: ProtocolEngine,
    hardware_api: HardwareAPI,
    task_queue: TaskQueue,
    python_and_legacy_file_reader: PythonAndLegacyFileReader,
    protocol_context_creator: ProtocolContextCreator,
    python_protocol_executor: PythonProtocolExecutor,
) -> PythonAndLegacyRunner:
    """Get a PythonAndLegacyRunner test subject with mocked dependencies."""
    return PythonAndLegacyRunner(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        task_queue=task_queue,
        python_and_legacy_file_reader=python_and_legacy_file_reader,
        protocol_context_creator=protocol_context_creator,
        python_protocol_executor=python_protocol_executor,
    )


@pytest.fixture
def live_runner_subject(
    protocol_engine: ProtocolEngine,
    hardware_api: HardwareAPI,
    task_queue: TaskQueue,
) -> LiveRunner:
    """Get a LiveRunner test subject with mocked dependencies."""
    return LiveRunner(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        task_queue=task_queue,
    )


@pytest.mark.parametrize(
    "config, runner_type",
    [
        (JsonProtocolConfig(schema_version=6), JsonRunner),
        (JsonProtocolConfig(schema_version=7), JsonRunner),
        (PythonProtocolConfig(api_version=APIVersion(2, 14)), PythonAndLegacyRunner),
        (JsonProtocolConfig(schema_version=5), PythonAndLegacyRunner),
        (PythonProtocolConfig(api_version=APIVersion(2, 13)), PythonAndLegacyRunner),
    ],
)
def test_create_protocol_runner(
    protocol_engine: ProtocolEngine,
    hardware_api: HardwareAPI,
    task_queue: TaskQueue,
    json_file_reader: JsonFileReader,
    json_translator: JsonTranslator,
    python_and_legacy_file_reader: PythonAndLegacyFileReader,
    protocol_context_creator: ProtocolContextCreator,
    python_protocol_executor: PythonProtocolExecutor,
    config: Union[JsonProtocolConfig, PythonProtocolConfig],
    runner_type: Type[AnyRunner],
) -> None:
    """It should return protocol runner type depending on the config."""
    assert isinstance(
        create_protocol_runner(
            protocol_config=config,
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
            task_queue=task_queue,
            json_file_reader=json_file_reader,
            json_translator=json_translator,
        ),
        runner_type,
    )


@pytest.mark.parametrize(
    "subject",
    [
        (lazy_fixture("json_runner_subject")),
        (lazy_fixture("python_runner_subject")),
        (lazy_fixture("live_runner_subject")),
    ],
)
def test_play_starts_run(
    decoy: Decoy,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    subject: AnyRunner,
) -> None:
    """It should start a protocol run with play."""
    subject.play(sentinel.deck_configuration)
    decoy.verify(
        protocol_engine.set_deck_configuration(sentinel.deck_configuration),
        protocol_engine.play(),
        times=1,
    )


@pytest.mark.parametrize(
    "subject",
    [
        (lazy_fixture("json_runner_subject")),
        (lazy_fixture("python_runner_subject")),
        (lazy_fixture("live_runner_subject")),
    ],
)
def test_pause(
    decoy: Decoy,
    protocol_engine: ProtocolEngine,
    subject: AnyRunner,
) -> None:
    """It should pause a protocol run with pause."""
    subject.pause()

    decoy.verify(protocol_engine.request_pause(), times=1)


@pytest.mark.parametrize(
    "subject",
    [
        (lazy_fixture("json_runner_subject")),
        (lazy_fixture("python_runner_subject")),
        (lazy_fixture("live_runner_subject")),
    ],
)
async def test_stop(
    decoy: Decoy,
    task_queue: TaskQueue,
    protocol_engine: ProtocolEngine,
    subject: AnyRunner,
) -> None:
    """It should halt a protocol run with stop."""
    decoy.when(protocol_engine.state_view.commands.has_been_played()).then_return(True)

    subject.play()
    await subject.stop()

    decoy.verify(await protocol_engine.request_stop(), times=1)


@pytest.mark.parametrize(
    "subject",
    [
        (lazy_fixture("json_runner_subject")),
        (lazy_fixture("python_runner_subject")),
        (lazy_fixture("live_runner_subject")),
    ],
)
async def test_stop_when_run_never_started(
    decoy: Decoy,
    task_queue: TaskQueue,
    protocol_engine: ProtocolEngine,
    subject: AnyRunner,
) -> None:
    """It should clean up rather than halt if the runner was never started."""
    decoy.when(protocol_engine.state_view.commands.has_been_played()).then_return(False)

    await subject.stop()

    decoy.verify(
        await protocol_engine.finish(
            drop_tips_after_run=False,
            set_run_status=False,
            post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
        ),
        times=1,
    )


@pytest.mark.parametrize(
    "subject",
    [
        (lazy_fixture("json_runner_subject")),
        (lazy_fixture("python_runner_subject")),
        (lazy_fixture("live_runner_subject")),
    ],
)
def test_resume_from_recovery(
    decoy: Decoy,
    protocol_engine: ProtocolEngine,
    subject: AnyRunner,
) -> None:
    """It should call `resume_from_recovery()` on the underlying engine."""
    subject.resume_from_recovery(
        reconcile_false_positive=sentinel.reconcile_false_positive
    )

    decoy.verify(
        protocol_engine.resume_from_recovery(
            reconcile_false_positive=sentinel.reconcile_false_positive
        ),
        times=1,
    )


async def test_run_json_runner(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    json_runner_subject: JsonRunner,
) -> None:
    """It should run a protocol to completion."""
    decoy.when(protocol_engine.state_view.commands.has_been_played()).then_return(
        False, True
    )

    assert json_runner_subject.was_started() is False
    await json_runner_subject.run(deck_configuration=sentinel.deck_configuration)
    assert json_runner_subject.was_started() is True

    decoy.verify(
        protocol_engine.set_deck_configuration(sentinel.deck_configuration),
        protocol_engine.play(),
        task_queue.start(),
        await task_queue.join(),
    )


# todo(mm, 2024-07-08): This test has grown long over time and it's unclear now what
# it's actually testing. Can we simplify or split it up?
async def test_run_json_runner_stop_requested_stops_enqueuing(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    json_runner_subject: JsonRunner,
    json_file_reader: JsonFileReader,
    json_translator: JsonTranslator,
) -> None:
    """It should run a protocol to completion."""
    labware_definition = LabwareDefinition2.model_construct()  # type: ignore[call-arg]
    json_protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=6),
        content_hash="abc123",
    )

    commands: List[pe_commands.CommandCreate] = [
        pe_commands.HomeCreate(params=pe_commands.HomeParams()),
        pe_commands.WaitForDurationCreate(
            params=pe_commands.WaitForDurationParams(seconds=10)
        ),
        pe_commands.LoadLiquidCreate(
            params=pe_commands.LoadLiquidParams(
                liquidId="water-id", labwareId="labware-id", volumeByWell={"A1": 30}
            )
        ),
    ]

    liquids: List[Liquid] = [
        Liquid(id="water-id", displayName="water", description="water desc")
    ]

    json_protocol = ProtocolSchemaV6.model_construct()  # type: ignore[call-arg]

    decoy.when(
        await protocol_reader.extract_labware_definitions(json_protocol_source)
    ).then_return([labware_definition])
    decoy.when(json_file_reader.read(json_protocol_source)).then_return(json_protocol)
    decoy.when(json_translator.translate_commands(json_protocol)).then_return(commands)
    decoy.when(json_translator.translate_liquids(json_protocol)).then_return(liquids)
    decoy.when(
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            pe_commands.HomeCreate(params=pe_commands.HomeParams()),
        )
    ).then_return(
        pe_commands.Home.model_construct(status=pe_commands.CommandStatus.SUCCEEDED)  # type: ignore[call-arg]
    )
    decoy.when(
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            pe_commands.WaitForDurationCreate(
                params=pe_commands.WaitForDurationParams(seconds=10)
            ),
        )
    ).then_return(
        pe_commands.WaitForDuration.model_construct(  # type: ignore[call-arg]
            id="protocol-command-id",
            error=pe_errors.ErrorOccurrence.from_failed(
                id="some-id",
                createdAt=datetime(year=2021, month=1, day=1),
                error=pe_errors.ProtocolEngineError(),
            ),
            status=pe_commands.CommandStatus.FAILED,
        )
    )
    decoy.when(
        protocol_engine.state_view.commands.get_error_recovery_type(
            "protocol-command-id"
        )
    ).then_return(ErrorRecoveryType.FAIL_RUN)

    await json_runner_subject.load(json_protocol_source)

    run_func_captor = matchers.Captor()

    decoy.verify(
        protocol_engine.add_labware_definition(labware_definition),
        protocol_engine.add_liquid(
            id="water-id", name="water", description="water desc", color=None
        ),
        protocol_engine.add_command(
            request=pe_commands.HomeCreate(params=pe_commands.HomeParams(axes=None))
        ),
        task_queue.set_run_func(func=run_func_captor),
    )

    # Verify that the run func calls the right things:
    run_func = run_func_captor.value

    with pytest.raises(pe_errors.ProtocolEngineError):
        await run_func()


@pytest.mark.filterwarnings("ignore::decoy.warnings.RedundantVerifyWarning")
@pytest.mark.parametrize(
    "schema_version, json_protocol",
    [
        (6, ProtocolSchemaV6.model_construct()),  # type: ignore[call-arg]
        (7, ProtocolSchemaV7.model_construct()),  # type: ignore[call-arg]
    ],
)
async def test_load_json_runner(
    decoy: Decoy,
    json_file_reader: JsonFileReader,
    json_translator: JsonTranslator,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    json_runner_subject: JsonRunner,
    schema_version: int,
    json_protocol: Union[ProtocolSchemaV6, ProtocolSchemaV7],
) -> None:
    """It should load a JSON protocol file."""
    labware_definition = LabwareDefinition2.model_construct()  # type: ignore[call-arg]

    json_protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=schema_version),
        content_hash="abc123",
    )

    commands: List[pe_commands.CommandCreate] = [
        pe_commands.WaitForResumeCreate(
            params=pe_commands.WaitForResumeParams(message="hello")
        ),
        pe_commands.WaitForResumeCreate(
            params=pe_commands.WaitForResumeParams(message="goodbye")
        ),
        pe_commands.LoadLiquidCreate(
            params=pe_commands.LoadLiquidParams(
                liquidId="water-id", labwareId="labware-id", volumeByWell={"A1": 30}
            )
        ),
    ]

    liquids: List[Liquid] = [
        Liquid(id="water-id", displayName="water", description="water desc")
    ]

    decoy.when(
        await protocol_reader.extract_labware_definitions(json_protocol_source)
    ).then_return([labware_definition])
    decoy.when(json_file_reader.read(json_protocol_source)).then_return(json_protocol)
    decoy.when(json_translator.translate_commands(json_protocol)).then_return(commands)
    decoy.when(json_translator.translate_liquids(json_protocol)).then_return(liquids)

    await json_runner_subject.load(json_protocol_source)

    run_func_captor = matchers.Captor()

    decoy.verify(
        protocol_engine.add_labware_definition(labware_definition),
        protocol_engine.add_liquid(
            id="water-id", name="water", description="water desc", color=None
        ),
        protocol_engine.add_command(
            request=pe_commands.HomeCreate(params=pe_commands.HomeParams(axes=None))
        ),
        task_queue.set_run_func(func=run_func_captor),
    )

    # Verify that the run func calls the right things:
    run_func = run_func_captor.value
    decoy.when(
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.WaitForResumeCreate(
                params=pe_commands.WaitForResumeParams(message="hello")
            ),
        )
    ).then_return(
        pe_commands.WaitForResume.model_construct(  # type: ignore[call-arg]
            id="command-id-1",
            status=CommandStatus.SUCCEEDED,
            error=None,
        )
    )
    decoy.when(
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.WaitForResumeCreate(
                params=pe_commands.WaitForResumeParams(message="goodbye")
            ),
        )
    ).then_return(
        pe_commands.WaitForResume.model_construct(  # type: ignore[call-arg]
            id="command-id-2",
            status=CommandStatus.SUCCEEDED,
            error=None,
        )
    )
    decoy.when(
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.LoadLiquidCreate(
                params=pe_commands.LoadLiquidParams(
                    liquidId="water-id", labwareId="labware-id", volumeByWell={"A1": 30}
                )
            ),
        )
    ).then_return(
        pe_commands.WaitForResume.model_construct(  # type: ignore[call-arg]
            id="command-id-3",
            status=CommandStatus.SUCCEEDED,
            error=None,
        )
    )
    await run_func()
    decoy.verify(
        # todo(mm, 2024-07-08): This triggers decoy.RedundantVerifyWarning.
        # Above, we're rehearsing each add_and_execute_command() call because the
        # subject needs the return values. But the subject doesn't return anything
        # based on on those return values, so here, we need to "redundantly" verify
        # each call to make sure the subject actually did some work.
        # It doesn't seem like Decoy wants subjects to be like this, so it's unclear
        # how to fix this.
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.WaitForResumeCreate(
                params=pe_commands.WaitForResumeParams(message="hello")
            ),
        ),
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.WaitForResumeCreate(
                params=pe_commands.WaitForResumeParams(message="goodbye")
            ),
        ),
        await protocol_engine.add_and_execute_command_wait_for_recovery(
            request=pe_commands.LoadLiquidCreate(
                params=pe_commands.LoadLiquidParams(
                    liquidId="water-id", labwareId="labware-id", volumeByWell={"A1": 30}
                )
            ),
        ),
    )


async def test_load_legacy_python(
    decoy: Decoy,
    python_and_legacy_file_reader: PythonAndLegacyFileReader,
    protocol_context_creator: ProtocolContextCreator,
    python_protocol_executor: PythonProtocolExecutor,
    task_queue: TaskQueue,
    protocol_engine: ProtocolEngine,
    python_runner_subject: PythonAndLegacyRunner,
) -> None:
    """It should load a legacy context-based Python protocol."""
    labware_definition = LabwareDefinition2.model_construct()  # type: ignore[call-arg]

    legacy_protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.py"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=PythonProtocolConfig(api_version=APIVersion(2, 11)),
        content_hash="abc123",
    )

    extra_labware = {"definition-uri": cast(LabwareDefinitionTypedDict, {})}

    legacy_protocol = PythonProtocol(
        text="",
        contents="",
        filename="protocol.py",
        api_level=APIVersion(2, 11),
        robot_type="OT-3 Standard",
        metadata={"foo": "bar"},
        bundled_labware=None,
        bundled_data=None,
        bundled_python=None,
        extra_labware=extra_labware,
    )

    protocol_context = decoy.mock(cls=ProtocolContext)

    decoy.when(
        await protocol_reader.extract_labware_definitions(legacy_protocol_source)
    ).then_return([labware_definition])
    decoy.when(
        python_and_legacy_file_reader.read(
            protocol_source=legacy_protocol_source,
            labware_definitions=[labware_definition],
            python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        )
    ).then_return(legacy_protocol)
    broker_captor = matchers.Captor()
    decoy.when(
        protocol_context_creator.create(
            protocol=legacy_protocol,
            broker=broker_captor,
            equipment_broker=matchers.IsA(Broker),
        )
    ).then_return(protocol_context)

    await python_runner_subject.load(
        legacy_protocol_source,
        python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        run_time_param_values=None,
        run_time_param_paths=None,
    )

    run_func_captor = matchers.Captor()

    decoy.verify(
        protocol_engine.add_labware_definition(labware_definition),
        protocol_engine.add_plugin(matchers.IsA(LegacyContextPlugin)),
        task_queue.set_run_func(run_func_captor),
    )

    assert broker_captor.value is python_runner_subject.broker

    # Verify that the run func calls the right things:
    run_func = run_func_captor.value
    await run_func()
    decoy.verify(
        await protocol_engine.add_and_execute_command(
            request=pe_commands.HomeCreate(params=pe_commands.HomeParams(axes=None))
        ),
        await python_protocol_executor.execute(
            protocol=legacy_protocol,
            context=protocol_context,
            run_time_parameters_with_overrides=None,
        ),
    )


async def test_load_python_with_pe_papi_core(
    decoy: Decoy,
    python_and_legacy_file_reader: PythonAndLegacyFileReader,
    protocol_context_creator: ProtocolContextCreator,
    protocol_engine: ProtocolEngine,
    python_runner_subject: PythonAndLegacyRunner,
) -> None:
    """It should load a legacy context-based Python protocol."""
    protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.py"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=PythonProtocolConfig(api_version=APIVersion(2, 14)),
        content_hash="abc123",
    )

    protocol = PythonProtocol(
        text="",
        contents="",
        filename="protocol.py",
        robot_type="OT-3 Standard",
        api_level=APIVersion(2, 14),
        metadata={"foo": "bar"},
        bundled_labware=None,
        bundled_data=None,
        bundled_python=None,
        extra_labware=None,
    )

    protocol_context = decoy.mock(cls=ProtocolContext)

    decoy.when(
        await protocol_reader.extract_labware_definitions(protocol_source)
    ).then_return([])
    decoy.when(
        python_and_legacy_file_reader.read(
            protocol_source=protocol_source,
            labware_definitions=[],
            python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        )
    ).then_return(protocol)
    broker_captor = matchers.Captor()
    decoy.when(
        protocol_context_creator.create(
            protocol=protocol, broker=broker_captor, equipment_broker=None
        )
    ).then_return(protocol_context)

    await python_runner_subject.load(
        protocol_source,
        python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        run_time_param_values=None,
        run_time_param_paths=None,
    )

    decoy.verify(protocol_engine.add_plugin(matchers.IsA(LegacyContextPlugin)), times=0)
    assert broker_captor.value is python_runner_subject.broker


async def test_load_legacy_json(
    decoy: Decoy,
    python_and_legacy_file_reader: PythonAndLegacyFileReader,
    protocol_context_creator: ProtocolContextCreator,
    python_protocol_executor: PythonProtocolExecutor,
    task_queue: TaskQueue,
    protocol_engine: ProtocolEngine,
    python_runner_subject: PythonAndLegacyRunner,
) -> None:
    """It should load a legacy context-based JSON protocol."""
    labware_definition = LabwareDefinition2.model_construct()  # type: ignore[call-arg]

    legacy_protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=5),
        content_hash="abc123",
    )

    legacy_protocol = JsonProtocol(
        text="{}",
        contents=cast(LegacyJsonProtocolDict, {}),
        filename="protocol.json",
        robot_type="OT-3 Standard",
        api_level=APIVersion(2, 11),
        schema_version=5,
        metadata={"protocolName": "A Very Impressive Protocol"},
    )

    protocol_context = decoy.mock(cls=ProtocolContext)

    decoy.when(
        await protocol_reader.extract_labware_definitions(legacy_protocol_source)
    ).then_return([labware_definition])
    decoy.when(
        python_and_legacy_file_reader.read(
            protocol_source=legacy_protocol_source,
            labware_definitions=[labware_definition],
            python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        )
    ).then_return(legacy_protocol)
    decoy.when(
        protocol_context_creator.create(
            legacy_protocol,
            broker=matchers.IsA(LegacyBroker),
            equipment_broker=matchers.IsA(Broker),
        )
    ).then_return(protocol_context)

    await python_runner_subject.load(
        legacy_protocol_source,
        python_parse_mode=PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS,
        run_time_param_values=None,
        run_time_param_paths=None,
    )

    run_func_captor = matchers.Captor()

    decoy.verify(
        protocol_engine.add_labware_definition(labware_definition),
        protocol_engine.add_plugin(matchers.IsA(LegacyContextPlugin)),
        task_queue.set_run_func(run_func_captor),
    )

    # Verify that the run func calls the right things:
    run_func = run_func_captor.value
    await run_func()
    decoy.verify(
        await protocol_engine.add_and_execute_command(
            request=pe_commands.HomeCreate(params=pe_commands.HomeParams(axes=None))
        ),
        await python_protocol_executor.execute(
            protocol=legacy_protocol,
            context=protocol_context,
            run_time_parameters_with_overrides=None,
        ),
    )


async def test_run_python_runner(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    python_runner_subject: PythonAndLegacyRunner,
) -> None:
    """It should run a protocol to completion."""
    decoy.when(protocol_engine.state_view.commands.has_been_played()).then_return(
        False, True
    )

    assert python_runner_subject.was_started() is False
    await python_runner_subject.run(deck_configuration=sentinel.deck_configuration)
    assert python_runner_subject.was_started() is True

    decoy.verify(
        protocol_engine.set_deck_configuration(sentinel.deck_configuration),
        protocol_engine.play(),
        task_queue.start(),
        await task_queue.join(),
    )


async def test_run_live_runner(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    protocol_engine: ProtocolEngine,
    task_queue: TaskQueue,
    live_runner_subject: LiveRunner,
) -> None:
    """It should run a protocol to completion."""
    decoy.when(protocol_engine.state_view.commands.has_been_played()).then_return(
        False, True
    )

    assert live_runner_subject.was_started() is False
    await live_runner_subject.run(deck_configuration=sentinel.deck_configuration)
    assert live_runner_subject.was_started() is True

    decoy.verify(
        await hardware_api.home(),
        protocol_engine.set_deck_configuration(sentinel.deck_configuration),
        protocol_engine.play(),
        task_queue.start(),
        await task_queue.join(),
    )



from opentrons.hardware_control import API, ThreadManager
from opentrons.types import Mount


async def test_synch_adapter() -> None:
    thread_manager = ThreadManager(API.build_hardware_simulator)
    synch = thread_manager.sync
    synch.cache_instruments({Mount.LEFT: "p10_single"})
    assert synch.attached_instruments[Mount.LEFT]["name"].startswith("p10_single")
    thread_manager.clean_up()



"""Camera interaction resource provider."""

import logging
from typing import Awaitable, Callable, Optional, Tuple

from pydantic import BaseModel, Field

from opentrons_shared_data.robot.types import RobotType

from ..errors import CameraCaptureError, CameraSettingsInvalidError

log = logging.getLogger(__name__)


class CameraError(BaseModel):
    """Generic base class for Camera errors that occur on entities handled through the Camera Provider."""

    message: str = Field(..., description="Description of error content.")
    code: str | None = Field(
        ..., description="Return code, if any, that was paired with the error."
    )


class CameraSettings(BaseModel):
    """Camera API settings for general enablement and use."""

    cameraEnabled: bool = Field(
        ..., description="Enablement status for general camera use."
    )
    liveStreamEnabled: bool = Field(
        ..., description="Enablement status for the Opentrons Live Stream service."
    )
    errorRecoveryCameraEnabled: bool = Field(
        ..., description="Enablement status for camera usage with Error Recovery."
    )


class ImageParameters(BaseModel):
    """Parameters for an Image Capture to determine filters. These are the inputs as expected by FFMPEG."""

    resolution: Optional[Tuple[int, int]] = Field(
        None,
        description="Width by height resolution in pixels for the image to be captured with.",
    )
    zoom: Optional[float] = Field(
        None,
        description="Multiplier to use when cropping and scaling a captured image.",
    )
    pan: Optional[Tuple[int, int]] = Field(
        None,
        description="Position to pan to for a given zoom. Format is X and Y coordinates (in pixels) to the bottom left of a frame.",
    )
    contrast: Optional[float] = Field(
        None, description="The contrast to use when processing an image."
    )
    brightness: Optional[int] = Field(
        None, description="The brightness to use when processing an image."
    )
    saturation: Optional[float] = Field(
        None, description="The saturation to use when processing an image."
    )


class CameraProvider:
    """Provider class to wrap camera interactions between the server and the engine."""

    def __init__(
        self,
        camera_settings_callback: Optional[Callable[[], CameraSettings]] = None,
        image_capture_callback: Optional[
            Callable[[RobotType, ImageParameters], Awaitable[bytes | CameraError]]
        ] = None,
    ) -> None:
        """Initialize the interface callbacks of the Camera Provider within the Protocol Engine.

        Params:
            camera_settings_callback: Callback to query the Camera Enablement settings from the Boolean Settings table.
            image_capture_callback: Callback to process an image capture request and return a bytestream of image data in response.
        """
        self._camera_settings_callback = camera_settings_callback
        self._image_capture_callback = image_capture_callback

    async def get_camera_settings(self) -> CameraSettings:
        """Query the Robot Server for the current Camera Enablement settings."""
        if self._camera_settings_callback is not None:
            return self._camera_settings_callback()
        # If we are in analysis or simulation, return as if the camera is enabled
        return CameraSettings(
            cameraEnabled=True, liveStreamEnabled=True, errorRecoveryCameraEnabled=True
        )

    async def capture_image(
        self, robot_type: RobotType, parameters: ImageParameters
    ) -> bytes | None:
        """Process through the Camera Executor on robot server an image capture request with a given set of filters.

        Returns a bytesteam of image data upon success. Raises an error if an error occurred during capture.
        Conditionally returns None if an image capture callback does not exist (simulation).
        """
        if self._image_capture_callback is not None:
            capture_result = await self._image_capture_callback(robot_type, parameters)
            if not isinstance(capture_result, CameraError):
                return capture_result
            else:
                if capture_result.code == "IMAGE_SETTINGS":
                    raise CameraSettingsInvalidError(message=capture_result.message)
                elif capture_result.code is not None:
                    error_str = f"Camera capture has failed to return an image with return code {capture_result.code}: {capture_result.message}"
                else:
                    error_str = f"Camera capture has failed with exception: {capture_result.message}"
                raise CameraCaptureError(message=error_str)
        # Return None if the image capture callback is unavailable (simulation)
        return None



"""Test Flex Stacker set stored labware command implementation."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.flex_stacker.set_stored_labware import (
    SetStoredLabwareImpl,
    SetStoredLabwareParams,
    SetStoredLabwareResult,
    StackerStoredLabwareDetails,
)
from opentrons.protocol_engine.errors import (
    FlexStackerNotLogicallyEmptyError,
)
from opentrons.protocol_engine.execution.equipment import (
    EquipmentHandler,
    LoadedLabwarePoolData,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    BatchLabwareLocationUpdate,
    BatchLoadedLabwareUpdate,
    FlexStackerPoolConstraint,
    FlexStackerStateUpdate,
    LabwareLidUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    InStackerHopperLocation,
    LabwareLocation,
    LabwareLocationSequence,
    LoadedLabware,
    NotOnDeckLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
    OverlapOffset,
    StackerStoredLabwareGroup,
)


@pytest.fixture
def subject(
    state_view: StateView, equipment: EquipmentHandler, model_utils: ModelUtils
) -> SetStoredLabwareImpl:
    """A FillImpl for testing."""
    return SetStoredLabwareImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


@pytest.mark.parametrize(
    "adapter_labware,lid_labware,pool_definition,initial_stored_labware,primary_loc_seq_prefixes,lid_loc_seq_prefixes,locations",
    [
        pytest.param(
            StackerStoredLabwareDetails(
                loadName="adapter-name", namespace="adapter-namespace", version=2
            ),
            StackerStoredLabwareDetails(
                loadName="lid-name", namespace="lid-namespace", version=3
            ),
            FlexStackerPoolConstraint(
                max_pool_count=10,
                pool_overlap=5.0,
                pool_height=sentinel.pool_height,
                primary_definition=sentinel.primary_definition,
                lid_definition=sentinel.lid_definition,
                adapter_definition=sentinel.adapter_definition,
            ),
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId="lid-id-2",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-3",
                    adapterLabwareId="adapter-id-3",
                    lidLabwareId="lid-id-3",
                ),
            ],
            [
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-2", lidId=None
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-3", lidId=None
                    )
                ],
            ],
            [
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-1", lidId="lid-id-1"
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    ),
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-2", lidId="lid-id-2"
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-2", lidId=None
                    ),
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-3", lidId="lid-id-3"
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-3", lidId=None
                    ),
                ],
            ],
            {
                "adapter-id-1": InStackerHopperLocation(moduleId="module-id"),
                "adapter-id-2": InStackerHopperLocation(moduleId="module-id"),
                "adapter-id-3": InStackerHopperLocation(moduleId="module-id"),
            },
            id="all-specified",
        ),
        pytest.param(
            None,
            None,
            FlexStackerPoolConstraint(
                max_pool_count=10,
                pool_overlap=5.0,
                pool_height=sentinel.pool_height,
                primary_definition=sentinel.primary_definition,
                lid_definition=None,
                adapter_definition=None,
            ),
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            [[], [], []],
            [[], [], []],
            {
                "labware-id-1": InStackerHopperLocation(moduleId="module-id"),
                "labware-id-2": InStackerHopperLocation(moduleId="module-id"),
                "labware-id-3": InStackerHopperLocation(moduleId="module-id"),
            },
            id="none-specified",
        ),
        pytest.param(
            None,
            StackerStoredLabwareDetails(
                loadName="lid-name", namespace="lid-namespace", version=3
            ),
            FlexStackerPoolConstraint(
                max_pool_count=10,
                pool_overlap=5.0,
                primary_definition=sentinel.primary_definition,
                lid_definition=sentinel.lid_definition,
                adapter_definition=None,
                pool_height=sentinel.pool_height,
            ),
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-1",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-1",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-2",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-2",
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-3",
                    adapterLabwareId=None,
                    lidLabwareId="lid-id-3",
                ),
            ],
            [[], [], []],
            [
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-1", lidId="lid-id-1"
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-2", lidId="lid-id-2"
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="labware-id-3", lidId="lid-id-3"
                    )
                ],
            ],
            {
                "labware-id-1": InStackerHopperLocation(moduleId="module-id"),
                "labware-id-2": InStackerHopperLocation(moduleId="module-id"),
                "labware-id-3": InStackerHopperLocation(moduleId="module-id"),
            },
            id="lid-only",
        ),
        pytest.param(
            StackerStoredLabwareDetails(
                loadName="adapter-name", namespace="adapter-namespace", version=2
            ),
            None,
            FlexStackerPoolConstraint(
                max_pool_count=10,
                pool_overlap=5.0,
                pool_height=sentinel.pool_height,
                primary_definition=sentinel.primary_definition,
                lid_definition=None,
                adapter_definition=sentinel.adapter_definition,
            ),
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-1",
                    adapterLabwareId="adapter-id-1",
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-2",
                    adapterLabwareId="adapter-id-2",
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-id-3",
                    adapterLabwareId="adapter-id-3",
                    lidLabwareId=None,
                ),
            ],
            [
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-2", lidId=None
                    )
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-3", lidId=None
                    )
                ],
            ],
            [[], [], []],
            {
                "adapter-id-1": InStackerHopperLocation(moduleId="module-id"),
                "adapter-id-2": InStackerHopperLocation(moduleId="module-id"),
                "adapter-id-3": InStackerHopperLocation(moduleId="module-id"),
            },
            id="adapter-only",
        ),
    ],
)
async def test_set_stored_labware_happypath(
    adapter_labware: StackerStoredLabwareDetails | None,
    lid_labware: StackerStoredLabwareDetails | None,
    pool_definition: FlexStackerPoolConstraint,
    decoy: Decoy,
    subject: SetStoredLabwareImpl,
    equipment: EquipmentHandler,
    state_view: StateView,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_adapter_def: LabwareDefinition,
    tiprack_lid_def: LabwareDefinition,
    initial_stored_labware: list[StackerStoredLabwareGroup],
    primary_loc_seq_prefixes: list[LabwareLocationSequence],
    lid_loc_seq_prefixes: list[LabwareLocationSequence],
    locations: dict[str, LabwareLocation],
) -> None:
    """It should load all possible main/lid/adapter combos."""
    module_id = "module-id"
    pool_definition.primary_definition = flex_50uL_tiprack
    pool_definition.lid_definition = (
        tiprack_lid_def if pool_definition.lid_definition is not None else None
    )
    pool_definition.adapter_definition = (
        tiprack_adapter_def if pool_definition.adapter_definition is not None else None
    )
    lid_definition: Any = None
    adapter_definition: Any = None
    params = SetStoredLabwareParams(
        moduleId=module_id,
        primaryLabware=StackerStoredLabwareDetails(
            loadName="main-name", namespace="main-namespace", version=1
        ),
        lidLabware=lid_labware,
        adapterLabware=adapter_labware,
        initialStoredLabware=initial_stored_labware,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        FlexStackerSubState(
            module_id=cast(FlexStackerId, module_id),
            pool_primary_definition=None,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            contained_labware_bottom_first=[],
            max_pool_count=0,
            pool_overlap=0,
            pool_height=0,
        )
    )
    decoy.when(
        await equipment.load_definition_for_details(
            load_name="main-name",
            namespace="main-namespace",
            version=1,
        )
    ).then_return((flex_50uL_tiprack, sentinel.unused))
    offset_ids_by_id: dict[str, str | None] = {}
    if lid_labware:
        decoy.when(
            await equipment.load_definition_for_details(
                load_name=lid_labware.loadName,
                namespace=lid_labware.namespace,
                version=lid_labware.version,
            )
        ).then_return((tiprack_lid_def, sentinel.unused))
        lid_definition = tiprack_lid_def
    if adapter_labware:
        decoy.when(
            await equipment.load_definition_for_details(
                load_name=adapter_labware.loadName,
                namespace=adapter_labware.namespace,
                version=adapter_labware.version,
            )
        ).then_return((tiprack_adapter_def, sentinel.unused))
        adapter_definition = tiprack_adapter_def

    for labware_group in initial_stored_labware:
        decoy.when(
            state_view.labware.known(labware_group.primaryLabwareId)
        ).then_return(True)
        offset_ids_by_id[labware_group.primaryLabwareId] = None
        if labware_group.adapterLabwareId:
            decoy.when(
                state_view.labware.known(labware_group.adapterLabwareId)
            ).then_return(True)
            decoy.when(
                state_view.geometry.get_location_sequence(
                    labware_group.adapterLabwareId
                )
            ).then_return(
                [
                    NotOnDeckLocationSequenceComponent(
                        logicalLocationName=OFF_DECK_LOCATION
                    )
                ]
            )
            decoy.when(
                state_view.labware.get_location(labware_group.primaryLabwareId)
            ).then_return(OnLabwareLocation(labwareId=labware_group.adapterLabwareId))
            decoy.when(
                state_view.labware.get_location(labware_group.adapterLabwareId)
            ).then_return(OFF_DECK_LOCATION)
            offset_ids_by_id[labware_group.adapterLabwareId] = None

        else:
            decoy.when(
                state_view.geometry.get_location_sequence(
                    labware_group.primaryLabwareId
                )
            ).then_return(
                [
                    NotOnDeckLocationSequenceComponent(
                        logicalLocationName=OFF_DECK_LOCATION
                    )
                ]
            )
            decoy.when(
                state_view.labware.get_location(labware_group.primaryLabwareId)
            ).then_return(OFF_DECK_LOCATION)

        if labware_group.lidLabwareId:
            decoy.when(
                state_view.labware.known(labware_group.lidLabwareId)
            ).then_return(True)
            decoy.when(
                state_view.labware.get_location(labware_group.lidLabwareId)
            ).then_return(OnLabwareLocation(labwareId=labware_group.primaryLabwareId))
            decoy.when(
                state_view.labware.get_lid_id_by_labware_id(
                    labware_group.primaryLabwareId
                )
            ).then_return(labware_group.lidLabwareId)
            offset_ids_by_id[labware_group.lidLabwareId] = None

    ordered_definitions = [
        x
        for x in [
            lid_definition,
            flex_50uL_tiprack,
            adapter_definition,
        ]
        if x is not None
    ]

    decoy.when(
        state_view.labware.stacker_labware_pool_to_ordered_list(
            primary_labware_definition=flex_50uL_tiprack,
            lid_labware_definition=lid_definition,
            adapter_labware_definition=adapter_definition,
        )
    ).then_return(ordered_definitions)

    decoy.when(
        state_view.geometry.get_height_of_labware_stack(ordered_definitions)
    ).then_return(sentinel.pool_height)

    decoy.when(
        state_view.labware.get_stacker_labware_overlap_offset(ordered_definitions)
    ).then_return(OverlapOffset(x=0, y=0, z=5.0))

    decoy.when(
        state_view.modules.stacker_max_pool_count_by_height(
            module_id, sentinel.pool_height, 5.0
        )
    ).then_return(10)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=SetStoredLabwareResult.model_construct(
            primaryLabwareDefinition=flex_50uL_tiprack,
            lidLabwareDefinition=lid_definition,
            adapterLabwareDefinition=adapter_definition,
            storedLabware=initial_stored_labware,
            count=3,
            originalPrimaryLabwareLocationSequences=[
                prefix
                + [
                    NotOnDeckLocationSequenceComponent(
                        logicalLocationName=OFF_DECK_LOCATION
                    )
                ]
                for prefix in primary_loc_seq_prefixes
            ],
            originalAdapterLabwareLocationSequences=(
                [
                    [
                        NotOnDeckLocationSequenceComponent(
                            logicalLocationName=OFF_DECK_LOCATION
                        )
                    ]
                    for _ in initial_stored_labware
                ]
                if initial_stored_labware[0].adapterLabwareId is not None
                else None
            ),
            originalLidLabwareLocationSequences=(
                [
                    prefix
                    + [
                        NotOnDeckLocationSequenceComponent(
                            logicalLocationName=OFF_DECK_LOCATION
                        )
                    ]
                    for prefix in lid_loc_seq_prefixes
                ]
                if initial_stored_labware[0].lidLabwareId is not None
                else None
            ),
            newPrimaryLabwareLocationSequences=[
                prefix + [InStackerHopperLocation(moduleId=module_id)]
                for prefix in primary_loc_seq_prefixes
            ],
            newAdapterLabwareLocationSequences=(
                [
                    [InStackerHopperLocation(moduleId=module_id)]
                    for _ in initial_stored_labware
                ]
                if initial_stored_labware[0].adapterLabwareId is not None
                else None
            ),
            newLidLabwareLocationSequences=(
                [
                    prefix + [InStackerHopperLocation(moduleId=module_id)]
                    for prefix in lid_loc_seq_prefixes
                ]
                if initial_stored_labware[0].lidLabwareId is not None
                else None
            ),
        ),
        state_update=StateUpdate(
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=module_id,
                pool_constraint=pool_definition,
                contained_labware_bottom_first=initial_stored_labware,
            ),
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id=locations,
                new_offset_ids_by_id=offset_ids_by_id,
            ),
        ),
    )


async def test_set_stored_labware_requires_empty_hopper(
    decoy: Decoy,
    state_view: StateView,
    subject: SetStoredLabwareImpl,
) -> None:
    """It should fail if the hopper is not empty."""
    module_id = "module-id"
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        FlexStackerSubState(
            module_id=cast(FlexStackerId, module_id),
            pool_primary_definition=None,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            contained_labware_bottom_first=[
                StackerStoredLabwareGroup(
                    primaryLabwareId="hello", adapterLabwareId=None, lidLabwareId=None
                )
            ],
            max_pool_count=6,
            pool_overlap=0,
            pool_height=0,
        )
    )
    with pytest.raises(FlexStackerNotLogicallyEmptyError):
        await subject.execute(
            SetStoredLabwareParams(
                moduleId=module_id,
                primaryLabware=StackerStoredLabwareDetails(
                    loadName="main-name", namespace="main-namespace", version=1
                ),
                lidLabware=None,
                adapterLabware=None,
                initialCount=3,
            )
        )


@pytest.mark.parametrize(
    "input_count,input_labware,output_labware,output_error",
    [
        (
            None,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
            does_not_raise(),
        ),
        (
            1,
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                )
            ],
            does_not_raise(),
        ),
    ],
)
@pytest.mark.parametrize(
    "overlap_override",
    [None, 5.0],
)
async def test_set_stored_labware_limits_count(
    input_count: int | None,
    input_labware: list[StackerStoredLabwareGroup] | None,
    output_labware: list[StackerStoredLabwareGroup],
    output_error: ContextManager[Any],
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: SetStoredLabwareImpl,
    model_utils: ModelUtils,
    flex_50uL_tiprack: LabwareDefinition,
    overlap_override: float | None,
) -> None:
    """It should default and limit the input count."""
    module_id = "module-id"
    params = SetStoredLabwareParams(
        moduleId=module_id,
        primaryLabware=StackerStoredLabwareDetails(
            loadName="opentrons_flex_96_filtertiprack_50ul",
            namespace="opentrons",
            version=1,
        ),
        lidLabware=None,
        adapterLabware=None,
        initialCount=input_count,
        initialStoredLabware=input_labware,
        poolOverlapOverride=overlap_override,
    )
    for i in range(len(output_labware)):
        decoy.when(model_utils.generate_id()).then_return(f"labware-{i + 1}")

    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        FlexStackerSubState(
            module_id=cast(FlexStackerId, module_id),
            pool_primary_definition=None,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            contained_labware_bottom_first=[],
            max_pool_count=0,
            pool_overlap=0,
            pool_height=0,
        )
    )
    decoy.when(
        await equipment.load_definition_for_details(
            load_name="opentrons_flex_96_filtertiprack_50ul",
            namespace="opentrons",
            version=1,
        )
    ).then_return(
        (flex_50uL_tiprack, "opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    )

    decoy.when(
        state_view.labware.stacker_labware_pool_to_ordered_list(
            flex_50uL_tiprack,
            None,
            None,
        )
    ).then_return([flex_50uL_tiprack])

    decoy.when(
        state_view.geometry.get_height_of_labware_stack([flex_50uL_tiprack])
    ).then_return(sentinel.pool_height)

    decoy.when(
        state_view.labware.get_stacker_labware_overlap_offset([flex_50uL_tiprack])
    ).then_return(OverlapOffset(x=0, y=0, z=10.0))

    decoy.when(
        state_view.modules.stacker_max_pool_count_by_height(
            module_id,
            sentinel.pool_height,
            10.0 if overlap_override is None else overlap_override,
        )
    ).then_return(2)

    # we need to control multiple return values from generate_id and it doesnt take
    # an argument so we can do this iter side-effecting thing
    labware_ids = iter(("labware-1", "labware-2"))
    decoy.when(model_utils.generate_id()).then_do(lambda: next(labware_ids))
    decoy.when(
        await equipment.load_labware_pool_from_definitions(
            pool_primary_definition=flex_50uL_tiprack,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            location=InStackerHopperLocation(moduleId=module_id),
            primary_id="labware-1",
            adapter_id=None,
            lid_id=None,
        )
    ).then_return(
        LoadedLabwarePoolData(
            primary_labware=LoadedLabware(
                id="labware-1",
                loadName="some-loadname",
                definitionUri="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
                location=InStackerHopperLocation(moduleId=module_id),
                lid_id=None,
                offsetId=None,
                displayName=None,
            ),
            adapter_labware=None,
            lid_labware=None,
        )
    )
    decoy.when(
        await equipment.load_labware_pool_from_definitions(
            pool_primary_definition=flex_50uL_tiprack,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            location=InStackerHopperLocation(moduleId=module_id),
            primary_id="labware-2",
            adapter_id=None,
            lid_id=None,
        )
    ).then_return(
        LoadedLabwarePoolData(
            primary_labware=LoadedLabware(
                id="labware-2",
                loadName="some-loadname",
                definitionUri="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
                location=InStackerHopperLocation(moduleId=module_id),
                lid_id=None,
                offsetId=None,
                displayName=None,
            ),
            adapter_labware=None,
            lid_labware=None,
        )
    )

    with output_error:
        result = await subject.execute(params)
    assert result == SuccessData(
        public=SetStoredLabwareResult.model_construct(
            primaryLabwareDefinition=flex_50uL_tiprack,
            lidLabwareDefinition=None,
            adapterLabwareDefinition=None,
            count=len(output_labware),
            storedLabware=output_labware,
            originalPrimaryLabwareLocationSequences=[
                [
                    NotOnDeckLocationSequenceComponent(
                        logicalLocationName=SYSTEM_LOCATION
                    )
                ]
                for _ in output_labware
            ],
            originalAdapterLabwareLocationSequences=None,
            originalLidLabwareLocationSequences=None,
            newPrimaryLabwareLocationSequences=[
                [InStackerHopperLocation(moduleId="module-id")] for _ in output_labware
            ],
            newAdapterLabwareLocationSequences=None,
            newLidLabwareLocationSequences=None,
        ),
        state_update=StateUpdate(
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=module_id,
                pool_constraint=FlexStackerPoolConstraint(
                    max_pool_count=2,
                    pool_overlap=overlap_override or 10.0,
                    pool_height=sentinel.pool_height,
                    primary_definition=flex_50uL_tiprack,
                    lid_definition=None,
                    adapter_definition=None,
                ),
                contained_labware_bottom_first=output_labware,
            ),
            batch_loaded_labware=BatchLoadedLabwareUpdate(
                new_locations_by_id={
                    f"labware-{i + 1}": InStackerHopperLocation(moduleId="module-id")
                    for i, _ in enumerate(output_labware)
                },
                offset_ids_by_id={
                    f"labware-{i + 1}": None for i, _ in enumerate(output_labware)
                },
                display_names_by_id={
                    f"labware-{i + 1}": None for i, _ in enumerate(output_labware)
                },
                definitions_by_id={
                    f"labware-{i + 1}": flex_50uL_tiprack
                    for i, _ in enumerate(output_labware)
                },
            ),
            labware_lid=LabwareLidUpdate(parent_labware_ids=[], lid_ids=[]),
        ),
    )


@pytest.mark.parametrize(
    "input_count,input_labware",
    [
        (3, None),
        (
            None,
            [
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-1",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-2",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
                StackerStoredLabwareGroup(
                    primaryLabwareId="labware-3",
                    adapterLabwareId=None,
                    lidLabwareId=None,
                ),
            ],
        ),
    ],
)
async def test_set_stored_labware_exceeding_max(
    input_count: int | None,
    input_labware: list[StackerStoredLabwareGroup] | None,
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: SetStoredLabwareImpl,
    model_utils: ModelUtils,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should default and limit the input count."""
    module_id = "module-id"
    params = SetStoredLabwareParams(
        moduleId=module_id,
        primaryLabware=StackerStoredLabwareDetails(
            loadName="opentrons_flex_96_filtertiprack_50ul",
            namespace="opentrons",
            version=1,
        ),
        lidLabware=None,
        adapterLabware=None,
        initialCount=input_count,
        initialStoredLabware=input_labware,
    )

    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        FlexStackerSubState(
            module_id=cast(FlexStackerId, module_id),
            pool_primary_definition=None,
            pool_adapter_definition=None,
            pool_lid_definition=None,
            contained_labware_bottom_first=[],
            max_pool_count=0,
            pool_overlap=0,
            pool_height=0,
        )
    )
    decoy.when(
        await equipment.load_definition_for_details(
            load_name="opentrons_flex_96_filtertiprack_50ul",
            namespace="opentrons",
            version=1,
        )
    ).then_return(
        (flex_50uL_tiprack, "opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    )

    decoy.when(
        state_view.labware.stacker_labware_pool_to_ordered_list(
            flex_50uL_tiprack,
            None,
            None,
        )
    ).then_return([flex_50uL_tiprack])

    decoy.when(
        state_view.labware.get_stacker_labware_overlap_offset([flex_50uL_tiprack])
    ).then_return(OverlapOffset(x=0, y=0, z=0))

    decoy.when(
        state_view.geometry.get_height_of_labware_stack([flex_50uL_tiprack])
    ).then_return(sentinel.pool_height)

    decoy.when(
        state_view.modules.stacker_max_pool_count_by_height(
            module_id, sentinel.pool_height, 0.0
        )
    ).then_return(2)

    with pytest.raises(CommandPreconditionViolated):
        await subject.execute(params)



import dataclasses

from typing_extensions import Self

from opentrons.types import Point


@dataclasses.dataclass(kw_only=True)
class AxisAlignedBoundingBox3D:
    """An axis-aligned bounding box (in other words, a cuboid in 3D space)."""

    min_x: float
    max_x: float

    min_y: float
    max_y: float

    min_z: float
    max_z: float

    @classmethod
    def from_corners(cls, corner_a: Point, corner_b: Point) -> Self:
        """Construct from two diagonally opposite corners.

        It doesn't matter which two corners, as long as they're diagonally opposite.
        It also doesn't matter what order.
        """
        return cls(
            min_x=min(corner_a.x, corner_b.x),
            max_x=max(corner_a.x, corner_b.x),
            min_y=min(corner_a.y, corner_b.y),
            max_y=max(corner_a.y, corner_b.y),
            min_z=min(corner_a.z, corner_b.z),
            max_z=max(corner_a.z, corner_b.z),
        )

    @property
    def x_dimension(self) -> float:
        """Return the dimension along the x-axis."""
        return self.max_x - self.min_x

    @property
    def y_dimension(self) -> float:
        """Return the dimension along the y-axis."""
        return self.max_y - self.min_y

    @property
    def z_dimension(self) -> float:
        """Return the dimension along the z-axis."""
        return self.max_z - self.min_z



"""Test seRailLights command."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.set_rail_lights import (
    SetRailLightsImplementation,
    SetRailLightsParams,
    SetRailLightsResult,
)
from opentrons.protocol_engine.execution import (
    RailLightsHandler,
)


async def test_set_rail_lights_implementation(
    decoy: Decoy,
    rail_lights: RailLightsHandler,
) -> None:
    """A setRailLights should have an execution implementation."""
    subject = SetRailLightsImplementation(rail_lights=rail_lights)

    data = SetRailLightsParams(
        on=True,
    )

    result = await subject.execute(data)

    assert result == SuccessData(public=SetRailLightsResult())

    decoy.verify(await rail_lights.set_rail_lights(True), times=1)



"""Task handling."""

from __future__ import annotations

import asyncio
import contextlib
import logging
from typing import AsyncIterator, Protocol

from opentrons_shared_data.errors.exceptions import EnumeratedError, PythonException

from ..actions import ActionDispatcher, FinishTaskAction, StartTaskAction
from ..errors import ErrorOccurrence
from ..resources import ConcurrencyProvider, ModelUtils
from ..state.state import StateStore
from ..types import Task

log = logging.getLogger(__name__)


class TaskFunction(Protocol):
    """The function run inside a task protocol."""

    async def __call__(self, task_handler: TaskHandler) -> None:
        """The function called inside a task."""
        ...


class TaskHandler:
    """Implementation logic for fast concurrency."""

    _state_store: StateStore
    _model_utils: ModelUtils
    _concurrency_provider: ConcurrencyProvider

    def __init__(
        self,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
        model_utils: ModelUtils | None = None,
        concurrency_provider: ConcurrencyProvider | None = None,
    ) -> None:
        """Initialize a TaskHandler instance."""
        self._state_store = state_store
        self._model_utils = model_utils or ModelUtils()
        self._concurrency_provider = concurrency_provider or ConcurrencyProvider()
        self._action_dispatcher = action_dispatcher

    async def create_task(
        self, task_function: TaskFunction, id: str | None = None
    ) -> Task:
        """Create a task and immediately schedules it."""
        task_id = self._model_utils.ensure_id(id)
        asyncio_task = asyncio.create_task(
            task_function(task_handler=self), name=f"engine-task-{task_id}"
        )

        def _done_callback(task: asyncio.Task[None]) -> None:
            try:
                maybe_exception = task.exception()
            except asyncio.CancelledError as e:
                maybe_exception = e
            if isinstance(maybe_exception, EnumeratedError):
                occurence: ErrorOccurrence | None = ErrorOccurrence.from_failed(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    error=maybe_exception,
                )
            elif isinstance(maybe_exception, BaseException):
                occurence = ErrorOccurrence.from_failed(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    error=PythonException(maybe_exception),
                )
            else:
                occurence = None
            try:
                self._action_dispatcher.dispatch(
                    FinishTaskAction(
                        task_id=task_id,
                        finished_at=self._model_utils.get_timestamp(),
                        error=occurence,
                    ),
                )
            except BaseException:
                log.exception("Exception in task finish dispatch.")

        asyncio_task.add_done_callback(_done_callback)
        task = Task(
            id=task_id,
            createdAt=self._model_utils.get_timestamp(),
            asyncioTask=asyncio_task,
        )
        self._action_dispatcher.dispatch(StartTaskAction(task))
        return task

    @staticmethod
    def _empty_queue(
        queue: "asyncio.Queue[asyncio.Task[None]]", this_task: asyncio.Task[None]
    ) -> None:
        """Empties the queue."""
        try:
            while True:
                task = queue.get_nowait()
                if task is this_task:
                    break
        except asyncio.QueueEmpty:
            pass

    @contextlib.asynccontextmanager
    async def synchronize_cancel_latest(self, group_id: str) -> AsyncIterator[None]:
        """Cancel current task."""
        lock = self._concurrency_provider.lock_for_group(group_id)
        if lock.locked():
            raise asyncio.CancelledError()
        async with lock:
            yield

    @contextlib.asynccontextmanager
    async def synchronize_cancel_previous(self, group_id: str) -> AsyncIterator[None]:
        """Cancel previous run."""
        queue = self._concurrency_provider.queue_for_group(group_id)
        while not queue.empty():
            task = queue.get_nowait()
            task.cancel()
        this_task = asyncio.current_task()
        assert this_task is not None
        queue.put_nowait(this_task)
        try:
            yield
        except asyncio.CancelledError:
            raise
        except BaseException:
            self._empty_queue(queue, this_task)
            raise
        else:
            self._empty_queue(queue, this_task)

    @contextlib.asynccontextmanager
    async def synchronize_sequential(self, group_id: str) -> AsyncIterator[None]:
        """Run tasks one after the other."""
        lock = self._concurrency_provider.lock_for_group(group_id)
        async with lock:
            yield

    @contextlib.asynccontextmanager
    async def synchronize_concurrent(self, group_id: str) -> AsyncIterator[None]:
        """Run a list of tasks at the same time."""
        yield

    def cancel_all(self, message: str | None = None) -> None:
        """Cancel all asyncio tasks immediately.

        Do not call this more than once synchronously because
        that could lead to tasks cancelling more than once.
        It can be called if there are no current tasks. In that case
        nothing will happen.
        """
        for task in self._state_store.tasks.get_all_current():
            task.asyncioTask.cancel(msg=message)
